Require Import List.
Require Import Arith.
Require Import step_defs.
Require Import Program.Equality.
Require Import Omega.
Require Import NPeano.

Hint Extern 2 =>
  match goal with
  | [ |- _ <= _ ] => omega
  | [ |- _ < _ ] => omega
  end.

Hint Extern 2 =>
  match goal with
  | [ H : context[?n + 1] |- _ ] => 
    rewrite (plus_comm n 1) in H; simpl in H
  | [ |- context[?n + 1] ] => 
    rewrite (plus_comm n 1); simpl
  end.

Ltac extract_open_lift :=
  repeat (match goal with
  | [ |- context[ exp_fun (lift ?a 1 2) ] ] =>
    replace (exp_fun (lift a 1 2)) with (lift (exp_fun a) 1 0) in *
         by (simpl; auto)
  | [ |- context[ exp_nrec (lift ?a ?n 2) ] ] =>
    replace (exp_nrec (lift a n 2)) with (lift (exp_nrec a) n 0) in *
         by (simpl; auto)
  | [ |- context[ exp_app (lift (exp_nrec ?a) 1 0) (exp_var 1) ] ] =>
    replace (exp_app (lift (exp_nrec a) 1 0) (exp_var 1))
       with (lift (exp_app (exp_nrec a) (exp_var 0)) 1 0) in *
         by (simpl; auto)
  | [ |- context[ exp_lam (lift ?a 1 1) ] ] =>
    replace (exp_lam (lift a 1 1)) with (lift (exp_lam a) 1 0) in *
         by (simpl; auto)
  | [ |- context[ exp_arr ({?n ~> ?b}?A) ({S ?n ~> lift ?b 1 0}?B)] ] =>
    replace (exp_arr ({n ~> b}A) ({S n ~> lift b 1 0}B))
       with ({n ~> b}(exp_arr A B)) by (simpl; auto)
  | [ H : context[ exp_arr ({?n ~> ?b}?A) ({S ?n ~> lift ?b 1 0}?B)] |- _ ] =>
    replace (exp_arr ({n ~> b}A) ({S n ~> lift b 1 0}B))
       with ({n ~> b}(exp_arr A B)) in H by (simpl; auto)
  | [ H : context[ exp_arr (lift ?A 1 0) (lift ?B 1 1) ] |- _ ] =>
    replace (exp_arr (lift A 1 0) (lift B 1 1))
       with (lift (exp_arr A B) 1 0) in H by (simpl; auto)
  | [ |- context[ exp_arr (lift ?A ?n ?m) (lift ?B ?n (S ?m))] ] =>
    replace (exp_arr (lift A n m) (lift B n (S m)))
       with (lift (exp_arr A B) n m) by (simpl; auto)
  | [ H : context[ exp_arr (lift ?A ?n ?m)  (lift ?B ?n (S ?m))] |- _ ] =>
    replace (exp_arr (lift A n m) (lift B n (S m)))
       with (lift (exp_arr A B) n m) in H by (simpl; auto)
  | [ |- context[ exp_sigma ({?n ~> ?b}?A) ({S ?n ~> lift ?b 1 0}?B)] ] =>
     replace (exp_sigma ({n ~> b}A) ({S n ~> lift b 1 0}B))
        with ({n ~> b}(exp_sigma A B)) by (simpl; auto)
  | [ H : context[ exp_sigma ({?n ~> ?b}?A) ({S ?n ~> lift ?b 1 0}?B)] |- _ ] =>
     replace (exp_sigma ({n ~> b}A) ({S n ~> lift b 1 0}B))
        with ({n ~> b}(exp_sigma A B)) in H by (simpl; auto)
  | [ |- context [exp_sigma (lift ?A1 ?n ?m) (lift ?A2 ?n (S ?m))] ] =>
      replace (exp_sigma (lift A1 n m) (lift A2 n (S m))) with
              (lift (exp_sigma A1 A2) n m) by auto
  | [ H : context [exp_sigma (lift ?A1 ?n ?m) (lift ?A2 ?n (S ?m))] |- _ ] =>
      replace (exp_sigma (lift A1 n m) (lift A2 n (S m))) with
              (lift (exp_sigma A1 A2) n m) in H by auto
  | [ |- context [ exp_mu ({S ?n ~> lift ?b 1 0}?A) ] ] =>
    replace (exp_mu ({S n ~> lift b 1 0}A))
       with ({n ~> b} (exp_mu A)) by (simpl; auto)
  | [ H : context [ exp_mu ({S ?n ~> lift ?b 1 0}?A) ] |- _ ] =>
    replace (exp_mu ({S n ~> lift b 1 0}A))
       with ({n ~> b} (exp_mu A)) in H by (simpl; auto)
  | [ |- context [ exp_eq ({?n ~> ?b}?a1) ({?n ~> ?b}?a2) ] ] =>
    replace (exp_eq ({n ~> b}a1) ({n ~> b}a2))
       with ({n ~> b}(exp_eq a1 a2)) by (simpl; auto)
  | [ H : context [ exp_eq ({?n ~> ?b}?a1) ({?n ~> ?b}?a2) ] |- _ ] =>
    replace (exp_eq ({n ~> b}a1) ({n ~> b}a2))
       with ({n ~> b}(exp_eq a1 a2)) in H by (simpl; auto)
  end).

Theorem plus_comm' : forall n m, m + n = n + m.
Proof. intros; rewrite plus_comm; auto. Qed.

Theorem plus_n_Sm' : forall n m, n + S m = (S n) + m.
Proof. intros; omega.  Qed.
Ltac swap_n_Sm := rewrite plus_n_Sm'.

Theorem pred_plus_n_Sm : forall n m, pred (n + S m) = n + m.
Proof.  intros; omega.  Qed.
Hint Rewrite pred_plus_n_Sm : step.


Hint Rewrite (plus_comm' 1) : step.
Hint Rewrite (plus_comm' 2) : step.
Hint Rewrite (plus_comm' 3) : step.

Ltac rewrite_ih :=
  repeat (match goal with
  | [ H : _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  | [ H : forall _ _ _ _ _ _ _ _, _ = _ |- _ ] => rewrite H by eauto; clear H
  end).

(* This file contains basic facts about the substitution and lift
   operations on terms.  It's all De Bruijn index wrangling *)
Ltac compares_cleanup :=
    match goal with
    | [ H : nat_compare _ _ = Eq |- _ ] => apply nat_compare_eq in H
    | [ H : nat_compare _ _ = Lt |- _ ] => apply nat_compare_lt in H
    | [ H : nat_compare _ _ = Gt |- _ ] => apply nat_compare_gt in H
    end.
  
Ltac destruct_compares n m :=
  remember (nat_compare n m);
  match goal with | [ H : ?comp = nat_compare n m |- _ ] =>
    apply eq_sym in H;
    destruct comp
  end;
  compares_cleanup.
  

Ltac clarify_compares :=
  match goal with
  | [ |- context[match nat_compare ?n ?m with
                 | Lt => _ | Eq => _ | Gt => _ 
                 end]
    ] => destruct_compares n m
  | [ H : ?P (match nat_compare ?n ?m with
              | Lt => _ | Eq => _ | Gt => _ 
              end) _ |- _
    ] => destruct_compares n m
  | [ H : ?P _ (match nat_compare ?n ?m with
                | Lt => _ | Eq => _ | Gt => _ 
                end) |- _
    ] => destruct_compares n m
  | [ H : context[(match nat_compare ?n ?m with
                | Lt => _ | Eq => _ | Gt => _ 
                end)] |- _
    ] => destruct_compares n m
  end.

Ltac db_var_case := 
  repeat clarify_compares; simpl in *; intuition;
  repeat clarify_compares; simpl in *; autorewrite with step in *; intuition.

Ltac invert_compare :=
  match goal with
  | [ H : match nat_compare _ _ with | Eq => _ | Lt => _ | Gt => _ end = _
     |- _ ] => clarify_compares; inversion H
  end.

Ltac destruct_typ_eq :=
  match goal with
  | [ H : {_ ~> _}?A = exp_nat |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~> _}?A = exp_arr _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H :  {_ ~> _}?A = exp_at _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~> _}?A = exp_mu _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~> _}?A = exp_eq _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~> _}?A = exp_sum _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  | [ H : {_ ~> _}?A = exp_sigma _ _ |- _ ] =>
    destruct A; simpl in H; inversion H; subst;
    try solve [simpl in *; invert_compare; simpl in *; auto]; clear H
  end.

Hint Rewrite <- plus_n_O : step.
Hint Rewrite app_length : step.

Ltac clarify_max :=
  repeat (match goal with
  | [ H : max _ _ <= _ |- _ ] =>
    apply Nat.max_lub_iff in H; destruct H
  | [ H : max _ _ < _ |- _ ] =>
    apply Nat.max_lub_lt_iff in H; destruct H
  end).

Ltac val_inversion :=
  repeat (match goal with
  | [ HV : val (exp_succ _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_roll _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_ncase _ _ _) |- _ ] => inversion HV
  | [ HV : val (exp_scase _ _ _) |- _ ] => inversion HV
  | [ HV : val (exp_pcase _ _) |- _ ] => inversion HV
  | [ HV : val (exp_inl _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_inr _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_pair _ _) |- _ ] => inversion HV; subst; clear HV
  | [ HV : val (exp_app _ _) |- _ ] => inversion HV
  | [ HV : val (exp_unroll _) |- _ ] => inversion HV
  end).

Ltac fo_inversion :=
  repeat (match goal with
  | [ HF : FO (exp_arr _ _) |- _ ] => inversion HF
  | [ HF : FO (exp_var _) |- _ ] => inversion HF
  | [ HF : FO (exp_mu _) |- _ ] => inversion HF
  | [ HF : FO (exp_sum _ _) |- _ ] => inversion HF; clear HF
  end); subst.

(* Opening and lifting closed terms/types does nothing *)
Theorem open_closed_id_gen : forall a n b,
  fvs a <= n -> {n ~> b}a = a.
Proof.
  induction a; intros; subst; simpl in *; 
  clarify_max; rewrite_ih; eauto; try db_var_case.
Qed.

Theorem open_closed_id : forall a n b,
  fvs a = 0 -> {n ~> b}a = a.
Proof.
  intros; eauto using open_closed_id_gen.
Qed.
Hint Resolve open_closed_id open_closed_id_gen.
Hint Rewrite open_closed_id_gen using (simpl; eauto 2; fail) : step.

Theorem lift_closed_id_gen : forall a n m,
  fvs a <= m -> lift a n m = a.
Proof.
  induction a; intros; subst; simpl in *; 
  clarify_max; rewrite_ih; eauto; try db_var_case.
Qed.

Theorem lift_closed_id : forall a n m,
  fvs a = 0 -> lift a n m = a.
Proof.
  intros; eauto using lift_closed_id_gen.
Qed.
Hint Rewrite lift_closed_id using (eauto 3; fail) : step.

Theorem lift_0_is_id : forall (b : exp) (m : nat),
  lift b 0 m = b.
Proof.
  induction b; intros; simpl in *; rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite lift_0_is_id : step.

(* Swapping and commuting lemmas *)
Theorem lift_lift_swap_gen : forall b m1 m2 n1 n2,
    m2 <= m1
  ->   lift (lift b n1 m1) n2 m2
     = lift (lift b n2 m2) n1 (m1 + n2).
Proof.
  induction b; intros; simpl; eauto; rewrite_ih; auto;
  db_var_case.
Qed.

Theorem lift_lift_swap_0 : forall b m1 n1 n2,
       lift (lift b n1 m1) n2 0
     = lift (lift b n2 0) n1 (n2 + m1).
Proof.
  intros; subst; rewrite lift_lift_swap_gen; eauto with arith.
Qed.
Hint Rewrite <- lift_lift_swap_0 : step.
Hint Extern 4 =>
  match goal with
  | [ H : context [ lift (lift ?A 1 1) ?n (S ?m) ]
     |- context [lift (lift ?A ?n ?m) 1 1] ] =>
    rewrite lift_lift_swap_gen with (b := A) by omega
  | [ |- context[ lift (exp_fun (lift _ 1 2)) _ (S ?m) ] ] =>
    extract_open_lift; replace (S m) with (1 + m) by auto;
    rewrite <- lift_lift_swap_0
  | [ |- context[ lift (exp_nrec (lift _ 1 2)) _ (S ?m) ] ] =>
    extract_open_lift; replace (S m) with (1 + m) by auto;
    rewrite <- lift_lift_swap_0
  | [ |- context[lift (lift _ _ ?m1) ?n2 0] ] =>
    rewrite lift_lift_swap_0
  | [ H : context[lift (lift _ _ ?m1) ?n2 0] |- _ ] =>
    rewrite lift_lift_swap_0 in H
  end.

Theorem lift_lift_collapse : forall b n1 n2 m1 m2,
     m1 <= m2
  -> m2 <= m1 + n1
  ->   lift (lift b n1 m1) n2 m2
     = lift b (n1 + n2) m1.
Proof.
  induction b; intros; simpl in *; auto; rewrite_ih; auto; db_var_case.
Qed.
Hint Rewrite lift_lift_collapse using omega : step.

(* Special case of the last theorem that comes up *)
Theorem lift_lift_collapse_1 : forall b n1,
       lift (lift b n1 0) 1 0
     = lift b (S n1) 0.
Proof.
  intros; autorewrite with step; simpl; auto.
Qed.
Hint Rewrite lift_lift_collapse_1 : step.

Theorem lift_open_commute : forall B b i n m,
     i <= m
  ->   lift ({i ~> b}B) n m
     = {i ~> lift b n m}(lift B n (S m)).
Proof.
  induction B; intros; simpl in *; rewrite_ih; auto; db_var_case.
Qed.
Hint Rewrite lift_open_commute using omega : step.
Hint Extern 4 =>
  match goal with
  | [ |- context [lift ({_ ~> _}_) _ _] ] =>
    rewrite lift_open_commute by (auto with arith)
  end.

Theorem open_lift_commute : forall b a n1 n2 i m,
     m <= n2
  -> m <= i 
  -> {n1 + i ~> (lift a (n1 + n2) 0)} (lift b n1 m) 
   = lift ({i ~> lift a n2 0} b) n1 m.
Proof.
  induction b; intros; simpl in *; eauto;
  autorewrite with step; simpl in *;
  repeat (match goal with
  | [ |- context [ S (_ + _) ] ] =>
    repeat (rewrite plus_n_Sm)
  | [ |- context [ (_ + _ + _) ] ] =>
    repeat (rewrite <- plus_assoc)
  end); rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite <- open_lift_commute using omega : step.

Theorem open_lift_commute_1 : forall b a n2 i m,
     m <= n2 -> m <= i
  ->   {S i ~> (lift a (S n2) 0)}(lift b 1 m)
     = lift ({i ~> lift a n2 0} b) 1 m.
Proof.
  intros b a n2 i m ? ?; replace (S i) with (1 + i) by auto;
  autorewrite with step in *; eauto.
Qed.
Hint Rewrite <- open_lift_commute_1 using omega : step.

Theorem open_lift_commute_0 : forall b a n1,
    {n1 ~> lift a n1 0} (lift b n1 0)
  = lift ({0 ~> a} b) n1 0.
Proof.
  intros b a n1; replace ({0 ~> a} b) with ({0 ~> lift a 0 0} b);
  try rewrite <- open_lift_commute by omega;
  autorewrite with step; eauto.
Qed.

Theorem open_lift_commute_1_closed : forall b a i m,
     m <= i
  -> fvs a = 0
  ->   {S i ~> a} (lift b 1 m) 
     = lift ({i ~> a} b) 1 m.
Proof.
  intros b a i m H1 HFV;
  replace ({i ~> a}) with ({i ~> lift a m 0})
       by (autorewrite with step; eauto);
  replace ({S i ~> a}) with ({S i ~> lift a (S m) 0})
       by (autorewrite with step; eauto);
  autorewrite with step; eauto.
Qed.
Hint Rewrite <- open_lift_commute_1_closed using (simpl; omega): step.

Theorem lift_ctx_length : forall n m G,
  length G = length (lift_c G n m).
Proof.
  induction G as [ | [? ?] ]; intros; simpl in *; eauto.
Qed.
Hint Rewrite <- lift_ctx_length : step.
Hint Resolve lift_ctx_length.

Theorem lift_ctx_app : forall G1 G2 n m,
    lift_c (G1 ++ G2) n m 
  = (lift_c G1 n (m + length G2)) ++ lift_c G2 n m.
Proof.
  induction G1 as [ | [? ?] ? ]; intros; simpl in *; eauto;
  try rewrite app_length; repeat (f_equal; try omega; eauto).
Qed.
Hint Rewrite lift_ctx_app : step.

Theorem open_under_lift : forall b a i n m,
     i <= n + m
  -> m <= i
  -> {i ~> a} (lift b (S n) m) = lift b n m.
Proof.
  induction b; intros; simpl in *; rewrite_ih; auto; try db_var_case.
Qed.
Hint Rewrite open_under_lift using omega : step.

Theorem open_into_lift : forall A n,
  {n ~> exp_var (S n)} (lift A 2 (S n)) = lift A 1 n.
Proof.
  induction A; intros; simpl; eauto;
  try solve [autorewrite with step; repeat rewrite_ih; eauto];
  db_var_case.
Qed.

Theorem open_open_swap : forall b n1 n2 a1 a2,
    n2 <= n1
  ->   {n1 ~> a1}({n2 ~> a2}b)
     = {n2 ~> ({n1 ~> a1}a2)}({S n1 ~> lift a1 1 n2}b).
Proof.
  induction b; intros; simpl in *; rewrite_ih; eauto;
  try match goal with
  | [ |- _ ] =>
    (* a little bit of fanciness so that we can see open_lift_commute applies *)
    f_equal; auto; match goal with
    | [ |- context[lift ({?i ~> ?a}?b) ?n 0] ] =>
      replace (lift ({i ~> a}b) n 0) with (lift ({i ~> lift a 0 0}b) n 0)
        by (rewrite lift_0_is_id; eauto)
    end; rewrite <- open_lift_commute; eauto
  end;
  try db_var_case.
Qed.

(* The de bruijn operations meet FO *)
Theorem lift_preserves_fo : forall A,
  FO A -> forall n m, FO (lift A n m).
Proof.
  induction 1; intros; simpl in *; eauto.
Qed.
Hint Resolve lift_preserves_fo.

Theorem open_preserves_fo : forall A,
  FO A -> forall a n, FO ({n ~> a} A).
Proof.
  induction 1; intros; simpl in *; eauto.
Qed.
Hint Resolve open_preserves_fo.

Theorem lifted_fo : forall A n m,
  FO (lift A n m) -> FO A.
Proof.
  induction A; inversion 1; subst; eauto.
Qed.
Hint Extern 1 =>
  match goal with
  | [ H : FO (lift ?A ?n ?m) |- FO ?A ] =>
    eapply lifted_fo; eexact H
  end.

Theorem pstep_preserves_fo : forall A,
  FO A -> forall B, pstep A B -> FO B.
Proof.
  induction 1; eauto; inversion 1; subst; eauto.
Qed.
Hint Immediate pstep_preserves_fo.

(* the de bruijn operations meet head_form *)
Theorem lift_preserves_hf : forall A h,
  head_form A h -> forall n m, head_form (lift A n m) h.
Proof.
  inversion 1; intros; simpl in *; subst; eauto.
Qed.
Hint Resolve lift_preserves_hf.

Theorem open_preserves_hf : forall A h,
  head_form A h -> forall n u, head_form ({n ~> u}A) h.
Proof.
  inversion 1; intros; simpl in *; subst; eauto.
Qed.
Hint Resolve open_preserves_hf.

(* The de bruijn operations meet reduction *)
Theorem lift_preserves_val : forall v,
  val v -> forall n m, val (lift v n m).
Proof.
  induction 1; intros; simpl in *; eauto.
Qed.
Hint Resolve lift_preserves_val.

Theorem open_preserves_val : forall v,
  val v -> forall v' n, val v' -> val ({n ~> v'} v).
Proof.
  induction 1; intros; simpl in *; eauto;
  clarify_compares; simpl; eauto.
Qed.
Hint Resolve open_preserves_val.

Theorem pstep_preserves_val : forall v v',
  pstep v v' -> val v -> val v'.
Proof.
  induction 1; intros; subst; val_inversion; eauto.
Qed.
Hint Immediate pstep_preserves_val.

Theorem lift_preserves_pstep : forall a b,
     pstep a b
  -> forall n m, 
       pstep (lift a n m) (lift b n m).
Proof.
  induction 1; intros; simpl in *; eauto;
  match goal with
  (* the natind case *)
  | [ |- pstep _ (lift ({0 ~> _} ({0 ~> exp_lam (exp_lam _)} _)) _ ?m) ] =>
    repeat (rewrite lift_open_commute by auto); simpl;
    replace (S (S (S (S (S m))))) with (S (S m) + 3) by omega;
    rewrite <- lift_lift_swap_gen by omega; eauto

  | [ |- pstep _ (lift ({0 ~> _} ({0 ~> _} ?a)) _ ?m) ] 
    => repeat (rewrite lift_open_commute by auto); eauto;
       extract_open_lift; replace (S m) with (1 + m) by (simpl; auto);
       rewrite <- lift_lift_swap_0; simpl; auto
                       
  end.
Qed.
Hint Resolve lift_preserves_pstep.

Theorem lift_preserves_mpstep : forall a b,
    mpstep a b
 -> forall n m,
      mpstep (lift a n m) (lift b n m).
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve lift_preserves_mpstep.

Theorem open_preserves_pstep_r :  forall a a',
  pstep a a' -> forall v n, val v -> pstep ({n ~> v}a) ({n ~> v}a').
Proof.
  induction 1; intros; simpl in *; eauto 7;
  rewrite open_open_swap with (n2 := 0) by omega; eauto;
  try (rewrite open_open_swap with (n2 := 0) by omega; eauto);
  simpl; autorewrite with step; simpl;
  try rewrite open_lift_commute_1 by omega; eauto 6.

(* The PCasePair case, sorry for not automating more. *)
rewrite open_open_swap with  (a2 := exp_join) (n2:=0) by omega;
rewrite lift_0_is_id; 
rewrite open_closed_id with (a := exp_join) by auto;
rewrite lift_lift_collapse by auto; auto.

(* natrec *)
  replace (S (S (S (S (S n))))) with (3 + (S (S n))) by omega.
  replace 5 with (3 + 2) by omega.
  rewrite open_lift_commute by omega.  eauto 6.
Qed.
Hint Resolve open_preserves_pstep_r.

Theorem open_preserves_mpstep_r : forall a a',
  mpstep a a' -> forall v n, val v -> mpstep ({n ~> v}a) ({n ~> v}a').
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve open_preserves_mpstep_r.

Theorem open_preserves_pstep_l : forall b a a',
  pstep a a' -> forall n, pstep ({n ~> a}b) ({n ~> a'}b).
Proof.
  induction b; intros; simpl in *; eauto 7;
  db_var_case.
Qed.
Hint Resolve open_preserves_pstep_l.

(* Variable numbers don't really matter *)
Theorem open_index_swap : forall a n1 n2,
  exists a', forall u, {n1 ~> u}a = {n2 ~> u}a'.
Proof.
  induction a; intros; simpl in *; eauto;
  try match goal with
  | [ |- context[nat_compare _ _] ] => 
    clarify_compares; subst; eauto;
    match goal with
    | [ |- exists _, forall u, u = {?n2 ~> u} _ ] =>
      exists (exp_var n2); db_var_case
    | [ |- exists _, forall _, exp_var ?n = {?n2 ~> _} _ ] =>
      destruct (le_gt_dec n2 n); 
        [ exists (exp_var (S n)); simpl; destruct n2; db_var_case
        | exists (exp_var n); simpl; db_var_case ]
    end
  | [ IHa : forall _ _, exists _, forall _, ({_ ~> _} ?a) = _,
      IHb : forall _ _, exists _, forall _, ({_ ~> _} ?b) = _ |-
      exists _, forall _, ?P ({_ ~> _} ?a) ({_ ~> _} ?b) = _]  =>
    let a' := fresh a in let b' := fresh b in
    edestruct IHa as [a' ?]; edestruct IHb as [b' ?];
    exists (P a' b'); simpl; intros; f_equal; eauto
  | [ IHa : forall _ _, exists _, forall _, ({_ ~> _} ?a) = _,
      IHb : forall _ _, exists _, forall _, ({_ ~> _} ?b) = _,
      IHc : forall _ _, exists _, forall _, ({_ ~> _} ?c) = _ |-
      exists _, forall _, ?P ({_ ~> _} ?a) ({_ ~> _} ?b) ({_ ~> _} ?c) = _] =>
    let a' := fresh a in let b' := fresh b in let c' := fresh c in
    edestruct IHa as [a' ?]; edestruct IHb as [b' ?]; edestruct IHc as [c' ?];
    exists (P a' b' c'); simpl; intros; f_equal; eauto
  | [ IH : forall _ _, exists _, forall _, ({_ ~> _} ?a) = _ |-
      exists _, forall _, ?P ({_ ~> _} ?a) = _]  =>
    let a' := fresh a in 
    edestruct IH as [a' ?];
    exists (P a'); simpl; intros; f_equal; eauto
  | [ IH : forall _ _, exists _, forall _, ({_ ~> _} ?a) = _ |-
      exists _, forall _, ?P ({_ ~> _} ?a) ?t = _]  =>
    let a' := fresh a in 
    edestruct IH as [a' ?];
    exists (P a' t); simpl; intros; f_equal; eauto
  | [ |- exists _, forall _, ?a = _ ] => 
    exists a; intros; eauto
  end.
Qed.

(* Lemmas about open_ctx *)
Theorem open_c_length : forall n v G,
  length (open_c n v G) = length G.
Proof.
  induction G as [ | [ ? ? ] ? ]; simpl in *; eauto.
Qed.
Hint Rewrite open_c_length : step.

Theorem open_c_app : forall n v G1 G2,
    open_c n v (G1 ++ G2) 
  = (open_c (n + length G2) v G1) ++ (open_c n v G2).
Proof.
  induction G1 as [ | [ ? ? ] ? ]; intros; simpl in *; eauto; 
  rewrite_ih; autorewrite with step in *; 
  repeat (f_equal; eauto; try omega).
Qed.
Hint Rewrite open_c_app : step.

Theorem open_open_swap_c : forall G a1 a2,
     fvs a1 = 0
  ->   open_c 0 a1 (open_c 0 a2 G)
     = open_c 0 ({0 ~> a1}a2) (open_c 1 a1 G).
Proof.
  induction G as [ | [? ?] ? ]; intros; simpl in *; auto;
  try solve
    [rewrite_ih; try rewrite <- open_lift_commute_0;
     rewrite open_open_swap by (autorewrite with step; eauto);
     autorewrite with step; eauto].
Qed.


