Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_reduction_basics.
Require Import step_typing_basics.
Require Import step_substitution.
Require Import step_preservation.
Require Import step_interp.
Require Import step_interp_facts.
Require Import step_confluence.

Require Import Program.Equality.

Ltac mstep_sval_elim :=
  match goal with
  | [ HS : mstep _ ?A _ |- _ ] =>
    let HV := fresh "HV" in assert (val A) as HV by constructor;
    destruct (val_mstep _ _ _ HV HS); subst; clear HV HS
  end.
(*

  | [ HS : mstep _ (exp_lam ?b) _ |- _ ] =>
    let HV := fresh "HV" in assert (val (exp_lam b)) as HV by auto;

  | [ HS : mstep _ (exp_fun ?b) _ |- _ ] =>
    let HV := fresh "HV" in assert (val (exp_fun b)) as HV by auto;
    destruct (val_mstep _ _ _ HV HS); subst; clear HV HS
  | [ HS : mstep _ exp_zero _ |- _ ] =>
    let HV := fresh "HV" in assert (val exp_zero) as HV by auto;
    destruct (val_mstep _ _ _ HV HS); subst; clear HV HS
  | [ HS : mstep _ (exp_arr ?A ?B) _ |- _ ] =>
    let HV := fresh "HV" in assert (val (exp_arr ?A ?B)) as HV by auto;
    destruct (val_mstep _ _ _ HV HS); subst; clear HV HS
  | [ HS : mstep _ exp_nat _ |- _ ] =>
    let HV := fresh "HV" in assert (val exp_nat) as HV by auto;
    destruct (val_mstep _ _ _ HV HS); subst; clear HV HS
  end.*)


Ltac mstep_inversion :=
  repeat (match goal with
  | [ HMS : mstep _ (exp_inl _) _ |- _ ] =>
    apply MSInlInversion in HMS; destruct HMS as [? [? HMS]]; subst;
    try match goal with
    | [ HEq : exp_inl _ = exp_inl _ |- _ ] => inversion HEq; subst; clear HEq
    end
  | [ HMS : mstep _ (exp_inr _) _ |- _ ] =>
    apply MSInrInversion in HMS; destruct HMS as [? [? HMS]]; subst;
    try match goal with
    | [ HEq : exp_inr _ = exp_inr _ |- _ ] => inversion HEq; subst; clear HEq
    end
  | [ HMS : mstep _ (exp_pair _ _) _ |- _ ] =>
    let HMS1 := fresh HMS "1" in let HMS2 := fresh HMS "2" in
    apply MSPairInversion in HMS;
    destruct HMS as [? [? [? [? [? [? [HMS1 HMS2]]]]]]]; subst;
    try match goal with
    | [ HEq : exp_pair _ _ = exp_pair _ _ |- _ ] => inversion HEq; subst; clear HEq
    end
  | [ HMS : mstep _ (exp_succ _) _ |- _ ] =>
    apply MSSuccInversion in HMS; destruct HMS as [? [? HMS]]
  | [ HMS : mstep _ (exp_roll _) _ |- _ ] =>
    apply MSRollInversion in HMS; destruct HMS as [? [? HMS]]
  | [ HV : val ?c, HMS : mstep _ (exp_unroll _) ?c |- _ ] =>
    apply mstep_unroll_val_inversion in HMS; auto;
    destruct HMS as [? [? HMS]]; subst
  | [ HV : val ?c, HMS : mstep _ (exp_ncase _ _ _) ?c |- _ ] =>
    apply mstep_ncase_val_inversion in HMS; auto;
    destruct_things
  | [ HV : val ?c, HMS : mstep _ (exp_scase _ _ _) ?c |- _ ] =>
    apply mstep_scase_val_inversion in HMS; auto;
    destruct_things
  | [ HV : val ?c, HMS : mstep _ (exp_pcase _ _) ?c |- _ ] =>
    apply mstep_pcase_val_inversion in HMS; auto;
    destruct_things
  | [ HV : val ?c, HMS : mstep _ (exp_app _ _) ?c |- _ ] =>
    apply mstep_app_val_inversion in HMS; auto;
    destruct_things; subst
  end; subst);
  repeat (mstep_sval_elim).

Ltac mstep_val_inversion :=
  mstep_inversion; val_inversion;
  try (match goal with
  | [ HMS1 : mstep _ ?a ?b, HMS2 : mstep _ ?a ?b' |- _ ] =>
    let HV1 := fresh "HV1" in let HV2 := fresh "HV2" in 
      assert (val b) as HV1 by eauto;
      assert (val b') as HV2 by eauto;
      destruct (mstep_val_deterministic _ _ _ _ _ HMS1 HMS2 HV1 HV2); 
      subst; clear HV1; clear HV2 
  end).


(* move these to the mstep facts file *)
Theorem MSPairCase1 : forall j a a' b,
  mstep j a a' -> mstep j (exp_pcase a b) (exp_pcase a' b).
Proof.
  induction 1; eauto.
Qed.

Ltac MSCongSolver := 
  repeat (instantiate; match goal with
  | [ |- mstep _ ?a ?a ] => apply MSRefl
  | [ H : mstep _ ?a ?b |- mstep _ ?a ?b ] => eapply H
  | [ H : mstep _ ?a ?b |- mstep _ (exp_app ?a _) _ ] =>
    eapply (MSTrans _ _ _ _ _ (MSApp1 _ _ _ _ H))
  | [ H : mstep _ ?a ?b |- mstep _ (exp_roll ?a) _ ] =>
    apply (MSRoll _ _ _ H)
  | [ H : mstep _ ?a ?b |- mstep _ (exp_inl ?a) _ ] =>
    apply (MSInl _ _ _ H)
  | [ H : mstep _ ?a ?b |- mstep _ (exp_inr ?a) _ ] =>
    apply (MSInr _ _ _ H)
  | [ Ha : mstep _ ?a ?a', Hb : mstep _ ?b ?b'
     |- mstep _ (exp_pair ?a ?b) _ ] =>
    let Hv := fresh "Hv" in
    assert (val a') as Hv by eauto;
    apply (MSPair _ _ _ _ _ _ Ha Hb Hv);
    clear Hv
  | [ H : mstep _ ?a ?b |- mstep _ (exp_succ ?a) _ ] =>
    apply (MSSucc _ _ _ H)
  | [ HAS : mstep _ ?a ?a' |- mstep _ (exp_ncase ?a ?b1 ?b2) _ ] =>
    apply MSTrans with (b := exp_ncase a' b1 b2);
    [ eauto using MSNatCase1 | ]
  | [ HAS : mstep _ ?a ?a' |- mstep _ (exp_scase ?a ?b1 ?b2) _ ] =>
    apply MSTrans with (b := exp_scase a' b1 b2);
    [ eauto using MSSumCase1 | ]
  | [ HAS : mstep _ ?a ?a' |- mstep _ (exp_pcase ?a ?b') _ ] =>
    apply MSTrans with (b := exp_pcase a' b');
    [ eauto using MSPairCase1 | ]
  | [ H : mstep _ ?b ?c |- mstep _ (exp_app ?a ?b) _ ] =>
    let HVal := fresh "HVal" in
    assert (val a) as HVal by eauto;
    eapply (MSTrans _ _ _ _ _ (MSApp2 _ _ _ _ H HVal)); clear HVal
  | [ H : interp _ _ _ ?v |- context[lift ?v _ _] ] =>
    rewrite (lift_closed_id v) by (eauto 3; fail)
  | [ |- context[ {0 ~> _} (close _ _ _) ] ] =>
      rewrite <- close_open_commute_closed by (eauto 2; fail)
  | [ |- mstep _ ?a _ ] =>
    try (eapply MSTrans; [ solve [ eapply MSStep; eauto ]; fail 2 | ])
  end).

Theorem minus_dist : forall m n p, n - (m + p) = n - m - p.
Proof.
  induction m; intros n p; destruct n; simpl; try omega.
Qed.

Ltac close_t_congs := 
  try (rewrite close_cong_arr by (eauto 3; fail));
  try (rewrite close_cong_mu by (eauto 3; fail));
  try (rewrite close_cong_sigma by (eauto 3; fail));
  try rewrite close_cong_nat; try rewrite close_cong_at;
  try rewrite close_cong_sum; try rewrite close_cong_eq;
  repeat rewrite close_open_commute by (eauto 3; fail).
Hint Extern 2 =>
  match goal with
  | [ H : typing _ _ _ ?A |- typing _ _ _ (close _ ?A _) ] =>
    close_t_congs
  | [ H : typing _ _ _ (?P _) |- typing _ _ _ (close _ (?P _) _) ] =>
    close_t_congs
  | [ H : typing _ _ _ (?P _ _) |- typing _ _ _ (close _ (?P _ _) _) ] =>
   close_t_congs
  | [ H : typing _ _ _ (?P _ _ _) |- typing _ _ _ (close _ (?P _ _ _) _) ] =>
    close_t_congs
  end.

Hint Extern 4 =>
match goal with
| [ H : ?j < ?j' |- interp (?j,_,_) _ _ _ ] =>
  apply downward_closure with (k := j')
| [ H : interp (?j1,?A,?cv) ?t ?rho ?a
   |- interp (?j2,?A,?cv) ?t ?rho ?a] =>
   try exact H; assert (j2 <= j1) by omega;
   apply downward_closure_leq with (k := j1); try assumption
end.

Theorem zero_in_interp : forall k t rho,
  interp (k, exp_nat, Comp) t rho exp_zero.
Proof.
  intros; unfold_interp;
  try match goal with
  | [ |- typing _ _ _ _ ] => interp_rewrite; eauto
  | [ |- exists _, exists _, _ /\ _ ] =>
    exists exp_zero; exists 0; split; try unfold_interp; eauto
  | [ |- forall _ _, _ -> _ -> _ -> _ ] =>
    intros; mstep_val_inversion; unfold_interp; eauto
  end.
Qed.
Hint Immediate zero_in_interp.

Theorem succ_in_interp : forall k t rho v,
     interp (k, exp_nat, Comp) t rho v
  -> interp (k, exp_nat, Comp) t rho (exp_succ v).
Proof.
  intros; unfold_interp_hyps; unfold_interp;
  try match goal with
  | [ |- typing _ _ _ _ ] => interp_rewrite; eauto
  | [ |- exists _, exists _, _ /\ _ ] =>
    eexists; eexists; split; [ MSCongSolver | eauto ]
  | [ |- forall _ _, _ -> _ -> _ -> _ ] =>
    intros; mstep_val_inversion; inst_interp; eauto
  end.
Qed.
Hint Resolve succ_in_interp.

(* move this to the interp_respects mpstep section *)
  Theorem interp_respects_mpstep_reduce : forall k A cv B1 t rho a G,
         mpstep (close rho B1 0) A
      -> typing G t B1 exp_star
      -> env_wf k G rho
      -> interp (k,B1,cv) t rho a
      -> interp (k,A,cv) t rho a.
  Proof.
    intros.
    replace rho with (nil ++ rho) by eauto.
    assert (typing nil t A exp_star)
        by (apply env_substitution_typ_all with (rho := rho) (k := k) in H0; eauto using mpstep_preservation).
    eapply interp_respects_mpstep with (B1 := B1) (G11 := nil) (G12 := nil) (G2 := G); eauto.
    apply typing_weakening_closed_type; eauto.
  Qed.


Section Soundness.
  Hint Resolve env_substitution_typ_all.

  Ltac inst_interp' :=
    interp_rewrite; repeat (match goal with
    | [ H1 : interp (?j, ?A, Val) ?t ?rho ?v,
        H2 : forall j _, j <= ?k -> interp (_, ?A, Val) ?t ?rho _ -> _ |- _ ] =>
      let Hle := fresh "Hle" in let HI := fresh "HI" in
      assert (j <= k) as Hle by eauto;
      assert (HI := H2 _ _ Hle H1); clear Hle H2; unfold_interp_hyps
    | [ H1 : interp (?j, ?A, Val) ?t ?rho ?v,
        H2 : forall j _, j < ?k -> interp (_, ?A, Val) ?t ?rho _ -> _ |- _ ] =>
      let Hle := fresh "Hle" in let HI := fresh "HI" in
      assert (j < k) as Hle by omega;
      assert (HI := H2 _ _ Hle H1); clear Hle H2; unfold_interp_hyps
    | [ H1 : mstep ?j ?a ?b,
        H2 : forall j _, j <= ?k -> mstep j ?a _ -> val _ -> interp _ _ _ _
       |- _ ] =>
      let Hle := fresh "Hle" in let HI := fresh "HI" in
      let HV := fresh "HV" in
      assert (j <= k) as Hle by eauto; assert (val b) as HV by eauto;
      assert (HI := H2 _ _ Hle H1 HV); clear Hle H2 HV; unfold_interp_hyps
    | [ IH : forall rho, 
                env_wf ?k (   (?t,(exp_eq (?P (?Q ?v')) (lift ?a 1 0)))
                           :: (?t',?A) :: ?G) rho
             -> interp _ _ _ _,
        HWf : env_wf ?k ?G ?rho, HMS : mstep _ (close ?rho ?a 0) (?P ?v) |- _ ] =>
      let HT := fresh "HT" in let IH' := fresh IH in try match goal with
      | [ HI : interp _ _ _ ?v |- _ ] => 
        assert (HT := interp_well_typed _ _ _ _ _ _ HI)
      end; assert (env_wf k (   (t,exp_eq (P (Q v')) (lift a 1 0))
                             :: (t',A) :: G)
                     (exp_join :: v :: rho)) as IH'
       by (apply EWFCons; [ | go_go_regularity; eauto | go_go_regularity; eauto];
           unfold_interp; rewrite open_under_lift,lift_0_is_id by eauto;
           interp_rewrite; go_go_regularity; autorewrite with step; eauto);
      apply IH in IH'; clear IH; rename IH' into IH;
      unfold_interp_hyps; try autorewrite with step in IH
    | [ IH : forall rho, 
                env_wf ?k (   (?t,(exp_eq (exp_pair ?v1' ?v2') (lift ?a 2 0)))
                           :: (?t',?A1) :: (?t',?A2) :: ?G) rho
             -> interp _ _ _ _,
        HWf : env_wf ?k ?G ?rho, HMS : mstep _ (close ?rho ?a 0) (exp_pair ?v1 ?v2) 
       |- _ ] => let IH' := fresh IH in 
      assert (env_wf k (   (t,exp_eq (exp_pair v1' v2') (lift a 2 0))
                        :: (t',A1) :: (t',A2) :: G)
                     (exp_join :: v2 :: v1 :: rho)) as IH'
          by (apply EWFCons; eauto; unfold_interp; 
              repeat (rewrite open_under_lift by eauto);
              rewrite lift_0_is_id; autorewrite with step;
              rewrite open_closed_id by eauto; eauto);
      apply IH in IH'; clear IH; rename IH' into IH;
      unfold_interp_hyps; try autorewrite with step in IH
    | [ Hv : val ?v,
        H2 : forall j v', j <= ?k -> mstep j (close ?rho ?v ?n) v' -> val _ -> interp _ _ _ _
       |- _ ] =>
      let Hle := fresh "Hle" in let HMS := fresh "HMS" in
      let HI := fresh "HI" in let HV := fresh "HV" in
      assert (0 <= k) as Hle by eauto;
      assert (mstep 0 (close rho v n) (close rho v n)) as HMS by eauto;
      assert (val (close rho v 0)) as HV by eauto;
      assert (HI := H2 _ _ Hle HMS HV); clear Hle H2 HMS HV; unfold_interp_hyps
    | [ IH : forall rho, env_wf ?k ((?t,?A) :: ?G) _
                      -> interp _ _ _ _,
        HWf : env_wf ?k ?G ?rho, HI : interp (_, ?A, _) ?t ?rho ?v
       |- _ ] =>
      let IH' := fresh IH in
      assert (env_wf k ((t,A) :: G) (inl _ v :: rho)) as IH'
        by (apply EWFCons; auto; try unfold_interp;
            autorewrite with step; eauto);
      apply IH in IH'; clear IH; rename IH' into IH;
      unfold_interp_hyps; try autorewrite with step in IH
    | [ IH : forall rho, env_wf ?k ((?t,exp_eq ?v ?a) :: ?G) _
                      -> interp _ _ _ _,
        IH2 : forall rho, env_wf _ ((_,exp_eq (exp_succ _) _) :: _) _
                      -> interp _ _ _ _,
        HWf : env_wf ?k ?G ?rho, HMS : mstep _ (close ?rho ?a 0) ?v
       |- _ ] =>
      let IH' := fresh IH in
      assert (env_wf k ((t,exp_eq v a) :: G)
                        (exp_join :: rho)) as IH' by
         (apply EWFCons; auto; try unfold_interp;
          autorewrite with step; eauto);
      apply IH in IH'; clear IH IH2; unfold_interp_hyps; rename IH' into IH;
      autorewrite with step in IH
    | [ H : interp (_, lift _ 1 0, _) _ (_ :: _) _ |- _ ] =>
      apply interp_strengthening_cons in H; [ | eauto 4; fail]
    | [ H : interp (_, lift _ 2 0, _) _ (_ :: _ :: _) _ |- _ ] =>
      apply interp_strengthening_cons_cons in H; [ | eauto 4; fail]
    | [ H : interp (_, lift _ 3 0, _) _ (_ :: _ :: _ :: _) _ |- _ ] =>
      apply interp_strengthening_cons_cons_cons in H; [ | eauto 4; fail]
    end).

  Lemma unfold_interp_eq : forall k b1 b2 cv rho p,
       interp (k, exp_eq b1 b2, cv) th_log rho p
    -> exists c, mpstep (close rho b1 0) c
              /\ mpstep (close rho b2 0) c.
  Proof.
    destruct cv; intros; unfold_interp_hyps; eauto.
  Qed.

  Hint Extern 0 =>
    match goal with
    | [ |- typing _ _ _ (close _ exp_star _) ] =>
      rewrite close_cong_star
    | [ H : typing ((th_log,exp_star) :: _) th_log ?A exp_star,
        HWF : env_wf ?k' _ ?rho'
       |- typing _ _ (exp_mu (close ?rho' ?A _)) _ ] =>
      apply TMu in H; apply env_substitution_typ_all with (rho := rho') (k := k') in H;
      apply_close_congs
    end.

  Hint Extern 0 => match goal with
  | [ |- typing _ _ (exp_fun (close _ _ _)) (close _ _ _) ] =>
    rewrite <- close_cong_fun by eauto
  | [ |- typing _ _ (exp_nrec (close _ _ _)) (close _ _ _) ] =>
    rewrite <- close_cong_nrec by eauto
  | [ |- typing _ _ (exp_ncase (close _ _ _) _ _) (close _ _ _) ] =>
    rewrite <- close_cong_ncase by eauto
  | [ |- typing _ _ (exp_inl (close _ _ _)) (close _ _ _) ] =>
    rewrite <- close_cong_inl
  | [ |- typing _ _ (exp_inr (close _ _ _)) (close _ _ _) ] =>
    rewrite <- close_cong_inr
  | [ |- typing _ _ (exp_scase (close _ _ _) _ _) (close _ _ _) ] =>
    rewrite <- close_cong_scase by eauto
  | [ |- typing _ _ (exp_pcase (close _ _ _) _) (close _ _ _) ] =>
    rewrite <- close_cong_pcase by eauto
  | [ |- typing _ _ (exp_pair (close _ _ _) (close _ _ _)) (close _ _ _) ] =>
    rewrite <- close_cong_pair by eauto
  | [ |- typing _ _ (exp_roll (close _ _ _)) (close _ _ _) ] =>
    rewrite <- close_cong_roll by eauto
  | [ |- typing _ _ (exp_unroll (close _ _ _)) (close _ _ _) ] =>
    rewrite <- close_cong_unroll by eauto
  | [ |- typing _ _ (exp_join) (close ?rho' _ ?m') ] =>
    rewrite <- close_cong_join with (rho := rho') (m := m') by eauto
  | [ H : typing _ _ (exp_sigma ?A ?B) _
     |- typing _ _ (exp_sigma (close _ ?A _) (close _ ?B _)) _ ] =>
     rewrite <- close_cong_sigma by (eauto 3; fail)
  end.

  Theorem nrec_fvs : forall t a A,
       typing nil t (exp_nrec a) A
    -> fvs a <= 2.
  Proof.
    intros t a A H; eapply typ_nrec_inversion_gen in H; [| eauto];
    destruct_things; match goal with
    | [ H : typing _ _ a _ |- _ ] => apply typing_fvs in H
    end; eauto.
  Qed.

  (* This theorem should really be generalized and moved elsewhere,
     but this overly specific version is enough for the one use case *)
  Theorem open_under_lift_lift : forall b a m,
         {S (S m) ~> a} (lift (lift b 1 (S m)) 1 m)
       = lift b 1 m.
  Proof.
    induction b; intros; simpl in *; rewrite_ih; auto;
    db_var_case.
  Qed.
  Hint Rewrite open_under_lift_lift using omega : step.

  Theorem typing_weakening_second_cons_star : forall t1 t2 B1 B2 G3 t a,
       typing ((t1,B1) :: G3) t a exp_star
    -> ctx_wf ((t2,B2) :: G3)
    -> fvs B1 = 0
    -> typing ((t1,B1) :: (t2,B2) :: G3) t 
              (lift a 1 1) 
              exp_star.
  Proof.
    intros;
    replace (exp_star) with (lift exp_star 1 1) in * by (simpl; auto);
    eauto using typing_weakening_second_cons.
  Qed.
  Hint Resolve typing_weakening_second_cons_star.

   Theorem close_lift_collapse : forall rho a n m,
        closed_rho rho
     -> fvs a <= m + length rho
     -> close rho (lift a n m) (n + m) = close rho a m.
   Proof.
     intros rho a n m HCR; revert a n m; induction HCR; intros.
     simpl in *; autorewrite with step in *; eauto using lift_closed_id_gen.
     simpl in *.
     replace a with (lift a (n + m) 0)
          by (autorewrite with step; eauto).
     rewrite open_lift_commute by eauto.
     repeat (rewrite (lift_closed_id a) by auto).  autorewrite with step.
     f_equal.
     replace (S (n + m)) with (n + (S m)) by omega.
     apply IHHCR.  omega.
   Qed.

   Theorem lift_close_collapse : forall rho a n m,
        closed_rho rho
     -> fvs a <= m + length rho
     -> lift (close rho a m) n m = close rho a m.
   Proof.
     intros rho a n m HCR; revert a n m; induction HCR; intros.
     simpl in *; autorewrite with step in *; eauto using lift_closed_id_gen.
     simpl in *; autorewrite with step in *; f_equal.
     apply IHHCR. omega.
   Qed.

   Theorem nrec_fvs_gen : forall G t a A,
        typing G t (exp_nrec a) A
     -> fvs a <= S (S (length G)).
   Proof.
     intros G t a A H; eapply typ_nrec_inversion_gen in H; [| eauto];
     destruct_things; match goal with
     | [ H : typing _ _ a _ |- _ ] => apply typing_fvs in H
     end; eauto.
   Qed.
   
   Theorem env_wf_length : forall k G rho,
      env_wf k G rho -> length G = length rho.
   Proof.
     induction 1; simpl in *; eauto.
   Qed.  Hint Immediate env_wf_length.

  Theorem interp_soundness : forall k G t a A,
       typing G t a A 
    -> forall rho, env_wf k G rho 
                -> interp (k,A,Comp) t rho (close rho a 0).
  Proof.
    induction k as [k IHWf] using lt_wf_ind;
    induction 1; intros; subst; eauto;
    try match goal with
    | [ |- interp (_, ?A, _) ?t _ ?a ] =>
      let HITyp := fresh "HITyp" in
      assert (typing nil t a (close_t rho A 0)) as HITyp by eauto;
      autorewrite with step in HITyp
    | [ HA : typing ?G ?t ?A exp_star, HFO : FO ?A,
        HB : typing ((_,?A)::_) _ ?B exp_star
       |- context[exp_sigma _ _] ] =>
      assert (typing G t (exp_sigma A B) exp_star) by eauto
    end; autorewrite with step;
    repeat (match goal with
    | [ Hwf : env_wf _ ?G _,
        IH : forall _, env_wf _ ?G _ -> interp _ _ _ _ |- _ ] =>
      let IH' := fresh IH in assert (IH' := IH _ Hwf); clear IH;
      rename IH' into IH
    end).

    Focus 13.

    (****** START APP ********)
    Hint Extern 0 => match goal with
    | [ |- typing _ _ (exp_app (close _ _ _) _) (close _ _ _) ] =>
      rewrite <- close_cong_app
    end.
    unfold_interp_hyps; unfold_interp; intros.
    Focus 3.

    (***** APP PROG *****)
    mstep_val_inversion; inst_interp'.
      (* PROG LAM *)

        replace (k - (1 + x + x0 + x1)) with (k - 1 - x - x0 - x1) 
             by (repeat (rewrite minus_dist); auto).
        apply downward_closure_leq with (j := k - 1 - x - x0) in HI0; [ | omega].
        assert (env_wf (k - 1 - x - x0 - x1) G rho) by 
          (eapply env_wf_downward_closure_leq; [eassumption |
            eapply le_trans; [ apply le_minus | eauto ]]).
        inst_interp'.
        go_go_regularity.
        assert (typing nil th_prog (close (x3 :: rho) B 0) exp_star) by
           (eapply env_substitution_typ_all with (G := (th_prog,A)::G); eauto).
        assert (k - 1 - x - x0 - x1 <= k - 1 - x - x0) by (apply le_minus).
       assert (interp (k - 1 - x - x0 -x1, close (x3 :: rho) B 0, Val) th_prog (x3 :: rho) v).
         eapply interp_respects_mpstep_reduce with (B1 := B) (G := (th_prog,A)::G).
          eauto. eauto. apply EWFCons.
          eapply downward_closure_leq; try eassumption; eauto.
          eauto. eauto. eauto.
        assert (mpstep (close (x3::rho) B 0) ({0 ~> x3}(close rho B 1))).
          simpl. rewrite close_open_commute by auto.
          rewrite close_closed_id by eauto.  eauto.
        assert (mpstep (close rho ({0 ~> b}B) 0) ({0 ~> x3}(close rho B 1))).
          rewrite close_open_commute by auto.  eauto using open_preserves_mpstep_l.
  apply interp_respects_mpstep_full with (A := {0 ~> x3}(close rho B 1))
                                         (B1 := close (x3 :: rho) B 0)
                                         (G := G).
  eauto. eauto.  eapply typing_weakening_closed_type; eauto. eauto.  eauto.
  apply interp_strengthening_cons with (v := x3). eauto.
    rewrite lift_closed_id by eauto. eauto.

     (* PROG FUN *)
        replace (k - (1 + x + x0 + x1)) with (k - 1 - x - x0 - x1) 
             by (repeat (rewrite minus_dist); auto).
        apply downward_closure_leq with (j := k - 1 - x - x0) in HI0; [ | omega].
        assert (env_wf (k - 1 - x - x0 - x1) G rho) by 
          (eapply env_wf_downward_closure_leq; [eassumption |
            eapply le_trans; [ apply le_minus | eauto ]]).
        inst_interp'.
        go_go_regularity.
        assert (typing nil th_prog (close (x3 :: rho) B 0) exp_star) by
           (eapply env_substitution_typ_all with (G := (th_prog,A)::G); eauto).
        assert (k - 1 - x - x0 - x1 <= k - 1 - x - x0) by (apply le_minus).
       assert (interp (k - 1 - x - x0 -x1, close (x3 :: rho) B 0, Val) th_prog (x3 :: rho) v).
         eapply interp_respects_mpstep_reduce with (B1 := B) (G := (th_prog,A)::G).
          eauto. eauto. apply EWFCons.
          eapply downward_closure_leq; try eassumption; eauto.
          eauto. eauto. eauto.
        assert (mpstep (close (x3::rho) B 0) ({0 ~> x3}(close rho B 1))).
          simpl. rewrite close_open_commute by auto.
          rewrite close_closed_id by eauto.  eauto.
        assert (mpstep (close rho ({0 ~> b}B) 0) ({0 ~> x3}(close rho B 1))).
          rewrite close_open_commute by auto.  eauto using open_preserves_mpstep_l.
  apply interp_respects_mpstep_full with (A := {0 ~> x3}(close rho B 1))
                                         (B1 := close (x3 :: rho) B 0)
                                         (G := G).
  eauto. eauto.  eapply typing_weakening_closed_type; eauto. eauto. eauto.
  apply interp_strengthening_cons with (v := x3). eauto.
    rewrite lift_closed_id by eauto. eauto.

      (* PROG NREC *)
        assert (fvs x2 <= 2) by eauto using nrec_fvs.
        rewrite (lift_closed_id_gen x2) in * by eauto.
        replace (k - (1 + x + x0 + x1)) with (k - 1 - x - x0 - x1) 
             by (repeat (rewrite minus_dist); auto).
        apply downward_closure_leq with (j := k - 1 - x - x0) in HI0; [ | omega].
        assert (env_wf (k - 1 - x - x0 - x1) G rho) by 
          (eapply env_wf_downward_closure_leq; [eassumption |
            eapply le_trans; [ apply le_minus | eauto ]]).
        inst_interp'.
        go_go_regularity.
        assert (typing nil th_prog (close (x3 :: rho) B 0) exp_star) by
           (eapply env_substitution_typ_all with (G := (th_prog,A)::G); eauto).
        assert (k - 1 - x - x0 - x1 <= k - 1 - x - x0) by (apply le_minus).
       assert (interp (k - 1 - x - x0 -x1, close (x3 :: rho) B 0, Val) th_prog (x3 :: rho) v).
         eapply interp_respects_mpstep_reduce with (B1 := B) (G := (th_prog,A)::G).
          eauto. eauto. apply EWFCons.
          eapply downward_closure_leq; try eassumption; eauto.
          eauto. eauto. eauto.
        assert (mpstep (close (x3::rho) B 0) ({0 ~> x3}(close rho B 1))).
          simpl. rewrite close_open_commute by auto.
          rewrite close_closed_id by eauto.  eauto.
        assert (mpstep (close rho ({0 ~> b}B) 0) ({0 ~> x3}(close rho B 1))).
          rewrite close_open_commute by auto.  eauto using open_preserves_mpstep_l.
  apply interp_respects_mpstep_full with (A := {0 ~> x3}(close rho B 1))
                                         (B1 := close (x3 :: rho) B 0)
                                         (G := G).
  eauto. eauto.  eapply typing_weakening_closed_type; eauto. eauto.  eauto.
  apply interp_strengthening_cons with (v := x3). eauto.
    rewrite lift_closed_id by eauto. eauto.
    (****** END APP ********)

 
    (* Ltac foo G IHWf k rho := *) try match goal with
       (* TESTED *)
    | [ |- interp (_, exp_at _ _, _) _ _ _ ] =>
      (* TBox cases *)
      unfold_interp_hyps; unfold_interp;
      match goal with
      | [ |- forall _ _, _ ] => intros; mstep_val_inversion; unfold_interp; eauto
        | [ H : val ?a |- exists _, exists _, _ /\ _ ] =>
          exists (close rho a 0); eexists; split; [ eauto | unfold_interp; eauto ]
        | [ H : mstep _ ?a ?v |- exists _, exists _, _ /\ _ ] =>
          exists v; eexists; split; [ eauto | unfold_interp; eauto using prog_contains_log ]
      end 
    | [ |- interp (_, exp_star, _) _ _ ?A ] =>
      (* All type formers *) 
      unfold_interp;
      try match goal with
      | [ |- exists _, exists _, _ /\ _ ] =>
        exists A; exists 0; split; [ eauto | unfold_interp; eauto]
      | [ |- forall _, _ ] =>
        intros; unfold_interp; mstep_val_inversion; eauto
      end
    | [ HHA : head_form ?A ?hA, HHB : head_form ?B ?hB,
        IH : interp (_, exp_eq ?A ?B, _) _ _ _|- _ ] =>
      (* TContra *)
      require_neq A B;
      apply unfold_interp_eq in IH; destruct IH as [C [MS1 MS2]];
      inversion HHA; subst; apply_close_congs; mpstep_inversion;
      inversion HHB; subst; apply_close_congs; mpstep_inversion;
      try match goal with
      | [ H : ?hf <> ?hf |- _ ] => destruct H; auto
      end
    | [ H : interp (_, exp_eq (?P _ _) (?P _ _), _) _ _ _
        |-  interp (_, exp_eq _ _, _) _ _ _ ] =>
      (* All equality inversions except mu*)
      unfold_interp_hyps; unfold_interp;
      try match goal with
      | [ H : mstep ?n ?A ?v |- exists _, exists _, mstep _ ?A _ /\ _ ] => 
        exists v; exists n; split; [ eauto | ]
      | [ |- forall _, _ ] => intros; inst_interp
      end; apply_close_congs; mpstep_inversion; unfold_interp;
      match goal with
      | [   H : mpstep (close ?rho ?B1 1) ?x
        |- exists _, mpstep (close ?rho ({0 ~> ?v}?B1) _) _ /\ _ ] =>
        repeat (rewrite close_open_commute by eauto 2);
        exists ({0 ~> close rho v 0}x); eauto 6 using open_preserves_mpstep_r
      | _ => eauto
      end
    | [ H : interp (_, exp_eq (exp_mu _) (exp_mu _), _) _ _ _
        |-  interp (_, exp_eq _ _, _) _ _ _ ] =>
      (* mu equality inversions *)
      unfold_interp_hyps; unfold_interp;
      try match goal with
      | [ H : mstep ?n ?A ?v |- exists _, exists _, mstep _ ?A _ /\ _ ] => 
        exists v; exists n; split; [ eauto | ]
      | [ |- forall _, _ ] => intros; inst_interp
      end; apply_close_congs; mpstep_inversion; unfold_interp;
      match goal with
      | [ H : mpstep (close ?rho ?A 1) ?x
         |- exists _, mpstep (close ?rho ({0 ~> exp_mu ?A} ?A) 0) _ /\ _ ] =>
        repeat (rewrite close_open_commute by eauto 2);
        exists ({0 ~> exp_mu x}x); apply_close_congs; split;
        eauto using open_preserves_mpstep, MPSMu
      end
    | [ |- interp _ _ _ (close _ (exp_var _) _) ] =>
      (* var case *)
      eauto using comp_contains_values
    | [ IH : interp (_, exp_at ?A _, _) _ _ ?v |- interp (_, ?A, _) _ _ ?v ] =>
      (* TUnboxVal *)
      unfold_interp; unfold_interp_hyps; intros; mstep_val_inversion;
      inst_interp'; eauto
    | [ H : interp (_, exp_eq ?b1 ?b2, _) _ _ _,
        HT : typing _ _ _ ({0 ~> ?b1}?A'),
        HTS : typing ?G' ?t ({0 ~> ?b2}?A') exp_star
       |- interp (_, {_ ~> _}?A', _) _ _ _ ] =>
      (* TConv *)
      apply unfold_interp_eq in H; destruct_things;
      go_go_regularity;
      match goal with
      | [ H : mpstep (close ?rho b1 _) ?x |- _ ] =>
        assert (typing G' t ({0 ~> b1}A') exp_star)
            by (apply regularity in HT; destruct_things; subst; auto;
                apply eq_star in HT; destruct_things; subst; auto;
                try star_isnt_equal);
        eapply interp_respects_mpstep_full with 
           (B1 := {0 ~> b1}A')
           (A := {0 ~> x}(close rho A' 1)) (G := G');
        try rewrite close_open_commute by eauto; eauto
      end
      (* UNTESTED *)
    (* nat cases *)
    | [ |- interp (_,exp_nat,_) _ _ _ ] => eauto
    (* FO case *)
    | [ H : FO ?A, HI : interp (_, ?A, _) th_prog _ _
       |- interp (_, ?A, _) th_log _ _ ] =>
      unfold_interp; unfold_interp_hyps; eauto 10
    (* subsumption case *)
    | [ H : interp ?p th_log ?rho ?a |- interp ?p th_prog ?rho ?a ] =>
      eauto using prog_contains_log
    (* Fun Case *)
    | [ H : env_wf _ ?G' ?rho
       |- interp (_, exp_arr ?B1 ?B2, Comp) th_prog ?rho
                 (exp_fun (close ?rho ?b _)) ] =>
      let HTyp1 := fresh "HTyp" in let HTyp2 := fresh "HTyp" in
      assert (typing G th_prog (exp_fun b) (exp_arr B1 B2)) as HTyp1 by eauto;
      assert (typing nil th_prog (exp_fun (close rho b 2)) (close rho (exp_arr B1 B2) 0))
          as HTyp2
          by (rewrite <- close_cong_fun by eauto; eapply env_substitution_nil; eauto);
      unfold_interp; intros; mstep_val_inversion; autorewrite with step;
      unfold_interp;
      match goal with
      | [ |- forall j v, j < _ -> interp _ _ _ v -> interp _ _ _ _ ] =>
        let j := fresh "j" in let v := fresh "v" in
        intros j v ? ?;  let IHApplied := fresh "IHApplied" in
        assert (env_wf j ((th_prog,lift (exp_arr B1 B2) 1 0)
                          :: (th_prog,B1) :: G')
                         (  (close (v :: rho) (lift (exp_fun b) 1 0) 0)
                          :: v :: rho)) as IHApplied by
          (apply EWFCons; [ | go_go_regularity; eauto | go_go_regularity; eauto];
           apply comp_interp_of_val; [eauto | ];
           apply IHWf with (G := (th_prog,B1) :: G); go_go_regularity; eauto);
        apply IHWf with (t := th_prog) (a := b) (A := lift B2 1 0) in IHApplied; auto;
        simpl in IHApplied; autorewrite with step in IHApplied;
        apply interp_strengthening_cons in IHApplied; eauto
      end
    (* Lam Case *)
    | [ H : typing ((?t,?A) :: ?G) ?t ?b ?B
       |- interp (_, exp_arr _ ?B, _) _ _ (exp_lam _) ] =>
      assert (typing G th_log (exp_lam b) (exp_arr A B)) by eauto;
      unfold_interp; intros; mstep_val_inversion;
      exists (exp_lam (close rho b 1)); exists 0; split; auto; unfold_interp;
      match goal with
      | [ |- forall _ _, _ <= _ -> _ -> _ ] =>
        let j := fresh "j" in let Hle := fresh "Hle" in
        intros j ? Hle; intros; go_go_regularity;
        destruct (le_lt_or_eq _ _ Hle); subst; eauto;
        match goal with
        | [ |- interp (?j,?B,Comp) th_log (?v :: ?rho)
                      ({0 ~> ?v} (close ?rho ?b 1)) ] =>
          assert (env_wf j G rho) by (eauto using env_wf_downward_closure_leq);
          rewrite <- close_open_commute_closed by eauto;
          assert (interp (j,B,Comp) th_log (v :: rho) (close (v :: rho) b 0)) by eauto;
          eauto
        end
      end
    (* Roll case *)
    | [ |- interp (_, exp_mu _, _) _ _ _ ] =>
      unfold_interp; unfold_interp_hyps;
      match goal with
      | [ H : mstep ?n (close _ _ _) ?v
         |- exists _, exists _, mstep _ _ _ /\ _ ] =>
        exists (exp_roll v); exists n; split; [ MSCongSolver | ];
        unfold_interp; interp_rewrite; eauto
      | [ |- forall _ _, _ -> _ -> _ -> _ ] =>
        intros; mstep_val_inversion; unfold_interp; unfold_interp_hyps;
        interp_rewrite; eauto 6
      end
    | [ IHtyp : interp (?k,_,_) _ _ (close ?rho ?a 0)
       |- interp (?k,_,_) _ _ (exp_inl (close ?rho ?a 0)) ] =>
      (* TInl, TInr *)
      unfold_interp; unfold_interp_hyps;
      match goal with
      | [ H : mstep _ ?a ?v |- exists _, exists _, _ /\ _ ] =>
        exists (exp_inl v); eexists; unfold_interp; split; [MSCongSolver | eauto]
      | [ |- forall _ _, _ -> mstep _ (exp_inl _) _ -> val _ -> _ ] =>
        intros ? ? ? HMS HVal; mstep_val_inversion; instantiate;
        subst; try (apply IHtyp in HMS); instantiate; auto;
        unfold_interp; eauto
      end
    | [ IHtyp : interp (?k,_,_) _ _ (close ?rho ?a 0)
       |- interp (?k,_,_) _ _ (exp_inr (close ?rho ?a 0)) ] =>
      (* TInl, TInr *)
      unfold_interp; unfold_interp_hyps;
      match goal with
      | [ H : mstep _ ?a ?v |- exists _, exists _, _ /\ _ ] =>
        exists (exp_inr v); eexists; unfold_interp; split; [MSCongSolver | eauto]
      | [ |- forall _ _, _ -> mstep _ (exp_inr _) _ -> val _ -> _ ] =>
        intros ? ? ? HMS HVal; mstep_val_inversion; instantiate;
        subst; try (apply IHtyp in HMS); instantiate; auto;
        unfold_interp; eauto
      end
    (* TJoin *)
    | [ |- interp (_, exp_eq _ _, Comp) th_log _ exp_join ] => 
      unfold_interp; exists exp_join; exists 0; split; auto;
      unfold_interp; eauto 7 using close_preserves_mpstep
    (* TUnroll *)
    | [ |- interp (_, {_ ~> _}_, _) ?t _ (exp_unroll _) ] =>
      unfold_interp_hyps; unfold_interp;
      interp_rewrite; eauto; intros; mstep_val_inversion; inst_interp'; eauto;
      match goal with
      | [ HI : forall _, _ < _ -> interp (_, {0 ~> exp_mu ?A}?A,Val) ?t ?rho ?v
         |- interp (?j, {0 ~> exp_mu ?A}?A, Val) th_prog ?rho ?v ] =>
        assert (interp (j, {0 ~> exp_mu A}A, Val) t rho v) by eauto;
        destruct t; eauto
      end
    (* sumcase case *)
    | [ |- interp (_, ?A, _) _ _ 
                  (exp_scase (close _ ?a _) (close _ ?b1 _)
                             (close _ ?b2 _)) ] =>
      unfold_interp; unfold_interp_hyps;
      match goal with
      | [ |- exists _, exists _, mstep _ _ _ /\ interp _ _ _ _ ] =>
        interp_rewrite; mstep_val_inversion; inst_interp';
        match goal with
        | [ H : interp (_, A, Val) _ _ ?v |- _ ] =>
          exists v; eexists; split; [ MSCongSolver | auto ]
        end
      | [ |- forall _ _, _ <= _ -> mstep _ _ _ -> val _ -> interp _ _ _ _ ] =>
        intros j v Hle HMSCase HVal; mstep_val_inversion; inst_interp';
        match goal with
        | [ HMSa : mstep ?j (close ?r _ _) (?P ?v),
            HT : typing ?G _ ?b _,
            HMSb : mstep _ ({_ ~> _} ({_ ~> _} (close _ ?b _))) _ |- _ ] =>
          remember (exp_join :: v :: r) as r';
          assert (env_wf (k - j - 1) G r') by
            (subst r'; apply EWFCons;
              [ unfold_interp; rewrite open_under_lift, lift_0_is_id by auto;
                [ apply_close_congs; eauto
                | autorewrite with step; eauto]
              | go_go_regularity; eauto using env_wf_downward_closure 
              | go_go_regularity; eauto using env_wf_downward_closure ]);
          eapply IHWf with (m := k - j - 1) (rho := r') in HT; subst r'; try omega; auto;
          unfold_interp_hyps; autorewrite with step in HT; inst_interp'; eauto
        end
      end
    | [ |- interp (_, ?B, _) _ _ 
                  (exp_pcase (close _ ?a _) (close _ ?b _)) ] =>
      unfold_interp; unfold_interp_hyps;
      try match goal with
      | [ |- exists _, exists _, mstep _ _ _ /\ interp _ _ _ _ ] =>
        go_go_regularity; interp_rewrite; mstep_val_inversion;
        inst_interp'; match goal with
        | [ H : interp (_, B, Val) _ _ ?v |- _ ] =>
          exists v; eexists; split; [ MSCongSolver | auto ]
        end
      | [ |- forall _ _, _ <= _ -> mstep _ _ _ -> val _ -> interp _ _ _ _ ] =>
        intros j v Hle HMSCase HVal; mstep_val_inversion; inst_interp';
        go_go_regularity;
        match goal with
        | [ HMSa : mstep ?j (close ?r _ _) (exp_pair ?va1 ?va2),
            HT : typing ?G _ ?b _,
            HMSb : mstep _ ({_ ~> _} ({_ ~> _} ({_ ~> _} (close _ ?b _)))) _
           |- _ ] => 
           remember (exp_join :: va2 :: va1 :: r) as r';
           assert (env_wf (k - j - 1) G r') by
             (subst r'; apply EWFCons; 
              [  unfold_interp; repeat (rewrite open_under_lift by eauto);
                   rewrite lift_0_is_id; autorewrite with step;
                   rewrite open_closed_id by eauto; eauto
               | eauto using env_wf_downward_closure | eauto]);
           eapply IHWf with (m := k - x - 1) (rho := r') in HT;
           subst r'; try omega; auto;
           unfold_interp_hyps; autorewrite with step in H0, HMSb;
           inst_interp'; eauto
        end
      end
    (* natcase case *)
    | [ |- interp (_,?A,_) ?t _ (exp_ncase (close _ ?a _) (close _ ?b1 _)
                                           (close _ ?b2 _)) ] =>
      assert (typing G t (exp_ncase a b1 b2) A) by eauto;
      unfold_interp; unfold_interp_hyps; try match goal with
      | [ H : looks_like_nat ?v
         |- exists v, exists n, _ /\ _ ] =>
        destruct v; simpl in H; try inversion H; subst; inst_interp';
        match goal with
        | [ HMS : mstep _ _ ?v, HI : interp _ _ _ ?v |- _ ] =>
          exists v; eexists; split; [ MSCongSolver | auto ]
        end
      | [ |-  forall _ _, _ <= _ -> _ -> _ -> interp _ _ _ _ ] =>
        intros; mstep_val_inversion; inst_interp'; eauto using downward_closure
      end
    end.

    (****** START APP ********)
    Hint Extern 0 => match goal with
    | [ |- typing _ _ (exp_app (close _ _ _) _) (close _ _ _) ] =>
      rewrite <- close_cong_app
    end.
    unfold_interp_hyps; unfold_interp; intros.

      (***** APP LAM ******)
    inst_interp'.
    go_go_regularity.

    assert (typing nil th_log (close (v1 :: rho) B 0) exp_star).
      eapply env_substitution_typ_all with (G := (th_log,A)::G) (k := k); eauto.

    assert (interp (k, close (v1 :: rho) B 0, Val) th_log (v1 :: rho) v3).
      apply interp_respects_mpstep_reduce with (B1 := B) (G := (th_log,A)::G) (k := k).
      eauto. eauto. eauto. eauto.

  assert (mpstep (close (v1::rho) B 0) ({0 ~> v1}(close rho B 1))).
    simpl.
    rewrite close_open_commute by eauto.  rewrite close_closed_id by eauto.  eauto.

  assert (mpstep (close rho ({0 ~> b}B) 0) ({0 ~> v1}(close rho B 1))).
    rewrite close_open_commute by eauto.  eauto using open_preserves_mpstep_l.

    match goal with
      | [ HI : interp (_, ?B, _) _ (_ :: _) ?v'
         |- exists v, exists _, mstep _ _ _ /\ interp (_, {_ ~> _} ?B, _) _ _ v ]
      =>  idtac v'
    end.

exists v3; eexists; split; [ MSCongSolver | idtac].

  apply interp_respects_mpstep_full with (A := {0 ~> v1}(close rho B 1))
                                         (B1 := close (v1 :: rho) B 0)
                                         (G := G).
  eauto. eauto.  eapply typing_weakening_closed_type; eauto. eauto. 
  eauto.

  apply interp_strengthening_cons with (v := v1). eauto.
    rewrite lift_closed_id by eauto. eauto.

    (**************** APP NREC **************)    
    inst_interp'.
    go_go_regularity.

    assert (typing nil th_log (close (v1 :: rho) B 0) exp_star).
      eapply env_substitution_typ_all with (G := (th_log,A)::G) (k := k); eauto.

    assert (interp (k, close (v1 :: rho) B 0, Val) th_log (v1 :: rho) v3).
      apply interp_respects_mpstep_reduce with (B1 := B) (G := (th_log,A)::G) (k := k).
      eauto. eauto. eauto. eauto.

  assert (mpstep (close (v1::rho) B 0) ({0 ~> v1}(close rho B 1))).
    simpl.
    rewrite close_open_commute by eauto.  rewrite close_closed_id by eauto.  eauto.

  assert (mpstep (close rho ({0 ~> b}B) 0) ({0 ~> v1}(close rho B 1))).
    rewrite close_open_commute by eauto.  eauto using open_preserves_mpstep_l.

  exists v3; eexists; split;
  [ MSCongSolver; rewrite lift_closed_id_gen by (eauto using nrec_fvs); eauto
  | idtac ].

  apply interp_respects_mpstep_full with (A := {0 ~> v1}(close rho B 1))
                                         (B1 := close (v1 :: rho) B 0)
                                         (G := G).
  eauto. eauto.  eapply typing_weakening_closed_type; eauto. eauto. 
  eauto.

  apply interp_strengthening_cons with (v := v1). eauto.
    rewrite lift_closed_id by eauto. eauto.

    (***** APP PROG *****)
    mstep_val_inversion; inst_interp'.
      (* PROG LAM *)

        replace (k - (1 + x + x0 + x1)) with (k - 1 - x - x0 - x1) 
             by (repeat (rewrite minus_dist); auto).
        apply downward_closure_leq with (j := k - 1 - x - x0) in HI0; [ | omega].
        assert (env_wf (k - 1 - x - x0 - x1) G rho) by 
          (eapply env_wf_downward_closure_leq; [eassumption |
            eapply le_trans; [ apply le_minus | eauto ]]).
        inst_interp'.
        go_go_regularity.
        assert (typing nil th_prog (close (x3 :: rho) B 0) exp_star) by
           (eapply env_substitution_typ_all with (G := (th_prog,A)::G); eauto).
        assert (k - 1 - x - x0 - x1 <= k - 1 - x - x0) by (apply le_minus).
       assert (interp (k - 1 - x - x0 -x1, close (x3 :: rho) B 0, Val) th_prog (x3 :: rho) v).
         eapply interp_respects_mpstep_reduce with (B1 := B) (G := (th_prog,A)::G).
          eauto. eauto. apply EWFCons.
          eapply downward_closure_leq; try eassumption; eauto.
          eauto. eauto. eauto.
        assert (mpstep (close (x3::rho) B 0) ({0 ~> x3}(close rho B 1))).
          simpl. rewrite close_open_commute by auto.
          rewrite close_closed_id by eauto.  eauto.
        assert (mpstep (close rho ({0 ~> b}B) 0) ({0 ~> x3}(close rho B 1))).
          rewrite close_open_commute by auto.  eauto using open_preserves_mpstep_l.
  apply interp_respects_mpstep_full with (A := {0 ~> x3}(close rho B 1))
                                         (B1 := close (x3 :: rho) B 0)
                                         (G := G).
  eauto. eauto.  eapply typing_weakening_closed_type; eauto. eauto.  eauto.
  apply interp_strengthening_cons with (v := x3). eauto.
    rewrite lift_closed_id by eauto. eauto.

     (* PROG FUN *)
        replace (k - (1 + x + x0 + x1)) with (k - 1 - x - x0 - x1) 
             by (repeat (rewrite minus_dist); auto).
        apply downward_closure_leq with (j := k - 1 - x - x0) in HI0; [ | omega].
        assert (env_wf (k - 1 - x - x0 - x1) G rho) by 
          (eapply env_wf_downward_closure_leq; [eassumption |
            eapply le_trans; [ apply le_minus | eauto ]]).
        inst_interp'.
        go_go_regularity.
        assert (typing nil th_prog (close (x3 :: rho) B 0) exp_star) by
           (eapply env_substitution_typ_all with (G := (th_prog,A)::G); eauto).
        assert (k - 1 - x - x0 - x1 <= k - 1 - x - x0) by (apply le_minus).
       assert (interp (k - 1 - x - x0 -x1, close (x3 :: rho) B 0, Val) th_prog (x3 :: rho) v).
         eapply interp_respects_mpstep_reduce with (B1 := B) (G := (th_prog,A)::G).
          eauto. eauto. apply EWFCons.
          eapply downward_closure_leq; try eassumption; eauto.
          eauto. eauto. eauto.
        assert (mpstep (close (x3::rho) B 0) ({0 ~> x3}(close rho B 1))).
          simpl. rewrite close_open_commute by auto.
          rewrite close_closed_id by eauto.  eauto.
        assert (mpstep (close rho ({0 ~> b}B) 0) ({0 ~> x3}(close rho B 1))).
          rewrite close_open_commute by auto.  eauto using open_preserves_mpstep_l.
  apply interp_respects_mpstep_full with (A := {0 ~> x3}(close rho B 1))
                                         (B1 := close (x3 :: rho) B 0)
                                         (G := G).
  eauto. eauto.  eapply typing_weakening_closed_type; eauto. eauto. eauto.
  apply interp_strengthening_cons with (v := x3). eauto.
    rewrite lift_closed_id by eauto. eauto.

      (* PROG NREC *)
        assert (fvs x2 <= 2) by eauto using nrec_fvs.
        rewrite (lift_closed_id_gen x2) in * by eauto.
        replace (k - (1 + x + x0 + x1)) with (k - 1 - x - x0 - x1) 
             by (repeat (rewrite minus_dist); auto).
        apply downward_closure_leq with (j := k - 1 - x - x0) in HI0; [ | omega].
        assert (env_wf (k - 1 - x - x0 - x1) G rho) by 
          (eapply env_wf_downward_closure_leq; [eassumption |
            eapply le_trans; [ apply le_minus | eauto ]]).
        inst_interp'.
        go_go_regularity.
        assert (typing nil th_prog (close (x3 :: rho) B 0) exp_star) by
           (eapply env_substitution_typ_all with (G := (th_prog,A)::G); eauto).
        assert (k - 1 - x - x0 - x1 <= k - 1 - x - x0) by (apply le_minus).
       assert (interp (k - 1 - x - x0 -x1, close (x3 :: rho) B 0, Val) th_prog (x3 :: rho) v).
         eapply interp_respects_mpstep_reduce with (B1 := B) (G := (th_prog,A)::G).
          eauto. eauto. apply EWFCons.
          eapply downward_closure_leq; try eassumption; eauto.
          eauto. eauto. eauto.
        assert (mpstep (close (x3::rho) B 0) ({0 ~> x3}(close rho B 1))).
          simpl. rewrite close_open_commute by auto.
          rewrite close_closed_id by eauto.  eauto.
        assert (mpstep (close rho ({0 ~> b}B) 0) ({0 ~> x3}(close rho B 1))).
          rewrite close_open_commute by auto.  eauto using open_preserves_mpstep_l.
  apply interp_respects_mpstep_full with (A := {0 ~> x3}(close rho B 1))
                                         (B1 := close (x3 :: rho) B 0)
                                         (G := G).
  eauto. eauto.  eapply typing_weakening_closed_type; eauto. eauto.  eauto.
  apply interp_strengthening_cons with (v := x3). eauto.
    rewrite lift_closed_id by eauto. eauto.
    (****** END APP ********)

    (*********   START NREC *********)
    
    assert (typing G th_log (exp_nrec a) (exp_arr exp_nat B)) by mycrush.

    assert (typing ((th_log, exp_nat) :: (th_log, exp_nat) :: G) th_log 
                   (exp_lam (exp_app (lift (exp_nrec a) 3 0)
                                              (exp_var 1)))
                   (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
                            (lift (lift B 1 1) 1 0)))
        as HTOneLam.
      go_go_regularity. eapply TLam; eauto.
      replace (exp_app (lift (exp_nrec a) 3 0) (exp_var 1))
         with (lift (exp_app (lift (exp_nrec a) 2 0) (exp_var 0))
                    1 0) by (simpl; autorewrite with step; auto).
      apply typing_weakening_cons; eauto.
      replace (lift (exp_nrec a) 2 0) with (lift (lift (exp_nrec a) 1 0) 1 1)
           by (autorewrite with step; eauto).
      replace (exp_app (lift (lift (exp_nrec a) 1 0) 1 1) (exp_var 0))
         with (lift (exp_app (lift (exp_nrec a) 1 0) (exp_var 0)) 1 1)
           by (simpl; eauto).
      apply typing_weakening_second_cons; eauto.
      rewrite <- (open_var_eq_gen B 0).
      apply TApp with (A := exp_nat).
        replace (exp_arr exp_nat (lift B 1 1))
           with (lift (exp_arr exp_nat B) 1 0)
             by eauto.
        apply typing_weakening_cons; eauto.
        eauto using TVar0_closed.
        rewrite open_var_eq_gen.  eauto.

    assert (typing ((th_log, exp_nat) :: G) th_log 
                   (exp_lam (exp_lam (exp_app (lift (exp_nrec a) 3 0)
                                              (exp_var 1))))
                   (exp_arr exp_nat 
                     (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
                              (lift (lift B 1 1) 1 0))))
         as HTTwoLam.
      go_go_regularity; eapply TLam; eauto.

   assert (env_wf k ((th_log,exp_nat)::G) (exp_zero::rho))
       by (constructor; try unfold_interp; eauto).

   assert (fvs (exp_lam (exp_lam
                 (exp_app (exp_nrec (close rho (lift a 2 2) 4))
                          (close rho (exp_var 1) 2)))) = 0).
      apply (env_substitution_nil _ _ _ _ _ _ H3) in HTTwoLam.
      apply typing_fvs_closed_exp in HTTwoLam.  simpl in HTTwoLam.
      apply_close_congs. apply_close_congs.
      rewrite open_under_lift in HTTwoLam by eauto.  eassumption.

    unfold_interp.
    exists (exp_nrec (close rho a 2)); exists 0; split; auto.
    unfold_interp.
    intros j v Hle HINat.
    
    assert (interp (k, B, Comp) th_log (v :: rho)
                   (exp_app (exp_nrec (close rho a 2)) v)).
      unfold_interp in HINat.  induction v; try inversion HINat.
      (***********  ZERO CASE **************)

      assert (fvs a <= S (S (length rho))) by 
          (go_go_regularity; replace (length rho) with (length G) by eauto;
                eapply nrec_fvs_gen; eauto).

      assert (env_wf k ((th_log,exp_nat)::G) (exp_zero::rho)) by eauto.
      assert (env_wf k 
        (   (th_log, exp_arr exp_nat
                      (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
                         (lift (lift B 1 1) 1 0)))
         :: (th_log, exp_nat) :: G)
        (   (close (exp_zero :: rho) 
               (lift (exp_lam (exp_lam (exp_app (exp_nrec (lift a 2 2))
                                                (exp_var 1)))) 1 0) 0)
         :: exp_zero :: rho)).
        apply EWFCons; [ | go_go_regularity; eauto | go_go_regularity; eauto].
        simpl.
        replace (exp_zero :: rho) with (close rho exp_zero 0 :: rho)
             by (autorewrite with step; eauto).
        eapply interp_rho_shift_out_cons;
          [ autorewrite with step; eassumption 
          | go_go_regularity; eauto
          | idtac ].
        simpl.  autorewrite with step.  

        (* Unfold first arrow *)
        unfold_interp.  
           apply env_substitution_nil
            with (k := k) (rho := exp_zero :: rho) in HTTwoLam.
           simpl in HTTwoLam; autorewrite with step in HTTwoLam;
           autorewrite with step; eauto. eassumption.
           
        intros.
        (* Unfold second arrow *)
        assert (fvs v = 0) by eauto.
        repeat (rewrite (lift_closed_id v) by eassumption).
        repeat (rewrite <- close_open_commute_closed by eauto).
        rewrite open_under_lift by eauto. simpl.

        assert (typing nil th_log
                       (exp_lam (exp_app (exp_nrec (close rho (lift a 1 2) 3))
                                         v))
                       (exp_arr (exp_eq (exp_succ v) exp_zero)
                                (close rho ({1 ~> v} (lift B 1 0)) 1))).
          apply env_substitution_nil
           with (k := k) (rho := v :: exp_zero :: rho) in HTOneLam.
          assert (fvs v = 0) by eauto.
          simpl in HTOneLam. autorewrite with step in HTOneLam.
          repeat (rewrite (lift_closed_id v) in * by eassumption).
          repeat (rewrite close_closed_id with (a := v) in * by eassumption).
          repeat (rewrite (open_closed_id v) in * by eassumption).
          rewrite <- close_open_commute_closed in HTOneLam by eauto.
          rewrite open_open_swap in HTOneLam by eauto.
          autorewrite with step in HTOneLam.
          repeat (rewrite (open_closed_id v) in * by eassumption).
          eassumption.  eauto.

        unfold_interp.
        autorewrite with step; rewrite (lift_closed_id v) by eauto;
        repeat (rewrite close_closed_id with (a := v) by eauto); eassumption.

        eexists; eexists; split.  eauto.
        unfold_interp.
        autorewrite with step; rewrite (lift_closed_id v) by eauto;
        repeat (rewrite close_closed_id with (a := v) by eauto); eassumption.
        intros.  unfold_interp in H12. apply_close_congs.
        mpstep_inversion.

      assert (IHtyping1' := IHtyping1 _ H7).  clear IHtyping1.
      eapply interp_rho_shift_in_cons in IHtyping1';
       [ | eauto | go_go_regularity; eauto ].
      simpl in IHtyping1'. autorewrite with step in IHtyping1'.

      unfold_interp.  autorewrite with step.  
      apply TApp with (A := exp_nat); go_go_regularity; eauto.
        apply env_substitution_nil with (k := k) (rho := rho) in H2;
        apply_close_congs; autorewrite with step in H2; eauto.
        rewrite <- close_open_commute_closed by eauto.
        apply env_substitution_nil with (k := k) (rho := exp_zero :: rho) in H19;
        apply_close_congs; autorewrite with step in H17; eauto.

      unfold_interp in IHtyping1'.
      eexists; eexists.  split.
      rewrite close_over_fvs_id with (a := exp_var 1) in Hastep
           by (simpl; eauto).
      MSCongSolver. rewrite lift_close_collapse by eauto.
      rewrite close_lift_collapse in Hastep by eauto.
      eassumption.  eauto.


      (***********  SUCC CASE **************)
      simpl in HINat.  assert (InnerIH := IHv HINat). clear IHv.
      assert (env_wf k ((th_log,exp_nat)::G) ((exp_succ v)::rho)) by eauto.
      assert (env_wf k 
        (   (th_log, exp_arr exp_nat
                      (exp_arr (exp_eq (exp_succ (exp_var 0)) (exp_var 1))
                         (lift (lift B 1 1) 1 0)))
         :: (th_log, exp_nat) :: G)
        (   (close ((exp_succ v) :: rho) 
               (lift (exp_lam (exp_lam (exp_app (exp_nrec (lift a 2 2))
                                                (exp_var 1)))) 1 0) 0)
         :: (exp_succ v) :: rho)).
        apply EWFCons; [ | go_go_regularity; eauto | go_go_regularity; eauto].
        simpl.
        replace ((exp_succ v) :: rho) with (close rho (exp_succ v) 0 :: rho)
             by (autorewrite with step; eauto).
        eapply interp_rho_shift_out_cons;
          [ autorewrite with step; eassumption 
          | go_go_regularity; eauto
          | idtac ].
        simpl.  autorewrite with step.  

        (* Unfold first arrow *)
        unfold_interp.  
           apply env_substitution_nil
            with (k := k) (rho := (exp_succ v) :: rho) in HTTwoLam.
           simpl in HTTwoLam; autorewrite with step in HTTwoLam;
           autorewrite with step; eauto.
           rewrite (lift_closed_id v) in HTTwoLam by eauto.
           eassumption. eassumption.
           
        intros.
        (* Unfold second arrow *)
        assert (fvs v0 = 0) by eauto.
        repeat (rewrite (lift_closed_id v0) by eassumption).
        repeat (rewrite <- close_open_commute_closed by eauto).
        rewrite open_under_lift by eauto. simpl.

        assert (typing nil th_log
                       (exp_lam (exp_app (exp_nrec (close rho (lift a 1 2) 3))
                                         v0))
                       (exp_arr (exp_eq (exp_succ v0) (exp_succ v))
                                (close rho ({1 ~> v0} (lift B 1 0)) 1))).
          apply env_substitution_nil
           with (k := k) (rho := v0 :: (exp_succ v) :: rho) in HTOneLam.
          assert (fvs v0 = 0) by eauto.
          simpl in HTOneLam. autorewrite with step in HTOneLam.
          repeat (rewrite (lift_closed_id v0) in * by eassumption).
          repeat (rewrite (lift_closed_id v) in * by eauto).
          autorewrite with step in HTOneLam.
          repeat (rewrite close_closed_id with (a := v0) in * by eassumption).
          repeat (rewrite close_closed_id with (a := v) in * by eauto).
          autorewrite with step in HTOneLam.
          rewrite <- close_open_commute_closed in HTOneLam by eauto.
          rewrite <- close_open_commute_closed in HTOneLam by eauto.
          rewrite open_open_swap in HTOneLam by eauto.
          autorewrite with step in HTOneLam.
          repeat (rewrite (open_closed_id v0) in * by eassumption).
          eassumption.  eauto.

        unfold_interp.
        autorewrite with step; rewrite (lift_closed_id v0) by eauto.
        repeat (rewrite close_closed_id with (a := v0) by eauto). 
        repeat (rewrite close_closed_id with (a := v) by eauto).
        rewrite open_closed_id by eauto. eassumption.

        eexists; eexists; split.  eauto.
        unfold_interp.
        autorewrite with step; rewrite (lift_closed_id v0) by eauto.
        repeat (rewrite close_closed_id with (a := v0) by eauto). 
        repeat (rewrite close_closed_id with (a := v) by eauto).
        rewrite open_closed_id by eauto. eassumption.

        intros.
        unfold_interp in H11. apply_close_congs. autorewrite with step in *.
        rewrite open_closed_id in H12 by eauto.
        mpstep_inversion.  unfold_interp in H7.

        Theorem mpstep_ll_nat : forall v v',
              looks_like_nat v
           -> mpstep v v'
           -> v = v'.
        Proof.
          induction v; try inversion 1; subst; intros. mpstep_inversion; eauto.
          mpstep_inversion; eauto.  rewrite IHv with (v' := x); eauto.
        Qed.

        apply mpstep_ll_nat in H11; [ | eauto].
        apply mpstep_ll_nat in H12; [ | eauto].
        inversion H11; subst; inversion H12; subst.
        apply interp_weakening_cons; [ | eauto].
        autorewrite with step.
        rewrite <- close_open_commute_closed by eauto.
        rewrite open_under_lift by eauto. 
        autorewrite with step.
        rewrite close_closed_id with (a := x0) by eauto.
        eauto using downward_closure.
      (* done wf assert *)

      assert (IHtyping1' := IHtyping1 _ H6).  clear IHtyping1.
      eapply interp_rho_shift_in_cons in IHtyping1';
       [ | eauto | go_go_regularity; eauto ].
      simpl in IHtyping1'.
      simpl in IHtyping1'; autorewrite with step in IHtyping1'.
  
      assert (fvs a <= S (S (length rho))) by 
          (go_go_regularity; replace (length rho) with (length G) by eauto;
                eapply nrec_fvs_gen; eauto).

      rewrite close_lift_collapse in  IHtyping1';
        [ | eauto | assumption ].
      
      unfold_interp.  autorewrite with step.  
      apply TApp with (A := exp_nat); go_go_regularity; eauto.
        apply env_substitution_nil with (k := k) (rho := rho) in H2;
        apply_close_congs; autorewrite with step in H2; eauto.
        rewrite <- close_open_commute_closed by eauto.
        apply env_substitution_nil with (k := k) (rho := (exp_succ v) :: rho) in H19;
        apply_close_congs; autorewrite with step in H19; eauto.
      unfold_interp in IHtyping1'.  exists v0. eexists.
      rewrite close_over_fvs_id with (a := exp_var 1) in Hastep by (simpl; eauto).
      
      split.  MSCongSolver. rewrite lift_closed_id_gen. eassumption.
      apply env_substitution_nil with (k := k) (rho := rho) in H2; 
        [ | eauto].  apply_close_congs.  eapply nrec_fvs; eauto.
      eauto.
    (** AFTER ASSERT **)
    unfold_interp in H5.
    unfold_interp.
      eapply mpreservation.  exact HTyp.  MSCongSolver.
      exists v0.
      apply mstep_app_val_inversion in Hastep.
      destruct_things.  
        mstep_val_inversion.  inversion H12.
        mstep_val_inversion.  inversion H12. 
        assert (val v) by eauto.
        mstep_val_inversion. inversion H13; subst.
        step_val_elim. exists x1.  eauto.  eauto.
     (*** end nrec ***)

    (****** START PAIR ********)
    match goal with
    | [|- interp (?k,exp_sigma ?A ?B,_) ?t ?r (exp_pair (close _ ?a _) (close _ ?b _)) ] =>
      (* Sigma *)
      assert (typing nil t (close rho (exp_pair a b) 0) (close rho (exp_sigma A B) 0))
        by eauto;
      unfold_interp_hyps; unfold_interp
    end.
    
      match goal with
      | [ HMSa : mstep ?n1 (close _ a _) ?va, HMSb : mstep ?n2 (close _ b _) ?vb
         |- exists _, exists _, _ /\ _ ] =>
        let HMSab := fresh "HMSab" in
        assert (mstep (n1 + n2) (close rho (exp_pair a b) 0) (exp_pair va vb))
          as HMSab by (autorewrite with step; MSCongSolver);
        exists (exp_pair va vb); eexists; split;
        [ autorewrite with step in HMSab; eauto | ]
      end.  
        unfold_interp. auto.

        apply interp_weakening_cons with (v := v1) in IHtyping4; [ | eauto].

        apply interp_respects_mpstep_full with 
          (A := close rho ({0 ~> v1} B) 0)
          (B1 := lift ({0 ~> a}B) 1 0)
          (G := (th_log,A) :: G); simpl; try rewrite open_under_lift,lift_0_is_id by eauto; auto.

          rewrite close_open_commute by eauto.
          rewrite close_open_commute_closed; eauto using open_preserves_mpstep_l.

      intros; unfold_interp; [rewrite <- close_cong_pair in H6; eauto | mstep_val_inversion].

        split; [ auto | split; [ auto | split; [inst_interp'; eauto | inst_interp' ]]].

        apply downward_closure_leq with (j := k - (x + x0)) in HI0; [ | eauto].
        apply downward_closure_leq with (j := k - (x + x0)) in HI; [ | eauto].
  
        apply interp_weakening_cons with (v := x1) in HI0; [ | eauto].
        apply interp_respects_mpstep_full with 
          (A := close rho ({0 ~> x1} B) 0)
          (B1 := lift ({0 ~> a}B) 1 0)
          (G := (th_prog,A) :: G); simpl; try rewrite open_under_lift,lift_0_is_id by eauto; auto.
  
        rewrite close_open_commute by eauto.
        rewrite close_open_commute_closed; eauto using open_preserves_mpstep_l.
  (* END PAIR *)

  (* START PCASE *)
Qed.
End Soundness.

Theorem normalization : forall a A,
     typing nil th_log a A
  -> exists v, exists n, val v /\ mstep n a v.
Proof.
  intros a A HTyp.
  apply interp_soundness with (k := 0) (rho := nil) in HTyp;
  try unfold_interp_hyps; eauto.
Qed.

Theorem mpstep_zero : forall a b,
  mpstep a b -> a = exp_zero -> b = exp_zero.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep exp_zero _ |- _ ] => 
    inversion H; subst; eauto
  end.
Qed.

Theorem mpstep_succ : forall a b,
  mpstep a b -> forall a', a = exp_succ a'
    -> exists b', b = exp_succ b'.
Proof.
  induction 1; intros; subst; eauto;
  match goal with
  | [ H : pstep (exp_succ _) _ |- _ ] => 
    inversion H; subst; eauto
  end.
Qed.

Theorem consistency : forall a,
     typing nil th_log a (exp_eq exp_zero (exp_succ exp_zero))
  -> False.
Proof.
  intros a HT;
  apply interp_soundness with (k := 0) (rho := nil) in HT; auto;
  unfold_interp_hyps.
  match goal with
  | [ HZ : mpstep exp_zero _, HS : mpstep (exp_succ _) _ |- _ ] =>
    apply mpstep_zero in HZ; auto;
    eapply mpstep_succ in HS; eauto;
    destruct HS; subst;
    match goal with
    | [ HContra : exp_zero = exp_succ _ |- _ ] => inversion HContra
    end
  end.
Qed.
