Require Import List.
Require Import step_defs.
Require Import step_db_basics.
Require Import step_typing_basics.
Require Import Program.Equality.
Require Import Omega.

Section Substitution.

  Variable v : exp.

  Theorem substitution_combined : 
       (   forall G t a A, typing G t a A
        -> forall G1 t' B G2, G = G1 ++ (t',B) :: G2
            -> val v
            -> typing G2 t' v B
            -> typing ((open_c 0 v G1) ++ G2) t
                      ({length G1 ~> lift v (length G1) 0} a)
                      ({length G1 ~> lift v (length G1) 0} A))
    /\ (   forall G, ctx_wf G
        -> forall G1 t' B G2, G = G1 ++ (t',B) :: G2
            -> val v
            -> typing G2 t' v B
            -> ctx_wf ((open_c 0 v G1) ++ G2)).
  Proof.
    apply typing_combined_ind;
    intros; subst; try match goal with
    | [ |- context[ exp_var (length _) ] ] =>
      (* TVar *)
      db_var_case; subst; simpl in *; invert_list_equalities; find_cons;
      inst_equalities; mycrush; eauto; simpl;
      match goal with
      | [ |- context[ ?n + S ?m ] ] =>
        replace (n + S m) with ((S n) + m) by omega
      end;
      [ rewrite open_lift_commute by omega | simpl; rewrite app_assoc];
      apply TVar; autorewrite with step; eauto
    | [ |- _ ] =>
      inst_equalities; mycrush; eauto 6;
      repeat (match goal with
      | [ |- context [{length ?n ~> _} ({0 ~> _} _)] ] =>
        rewrite open_open_swap with (n1 := length n) by omega
      | [ H : context [{length ?n ~> _} ({0 ~> _} _)] |- _ ] =>
        rewrite open_open_swap with (n1 := length n) in H by omega
      end); mycrush; eauto
    end.
  Qed.

  Theorem typing_substitution : forall G1 t' B G2 t a A,
      val v
    -> typing G2 t' v B
    -> typing (G1 ++ (t',B) :: G2) t a A
    -> typing ((open_c 0 v G1) ++ G2) t
              ({length G1 ~> lift v (length G1) 0} a)
              ({length G1 ~> lift v (length G1) 0} A).
  Proof.
    intros; destruct substitution_combined as [SUB ?]; eauto.
  Qed.
  
  Theorem ctx_wf_substitution : forall G1 t' B G2,
       ctx_wf (G1 ++ (t',B) :: G2)
    -> val v
    -> typing G2 t' v B
    -> ctx_wf ((open_c 0 v G1) ++ G2).
  Proof.
    intros; destruct substitution_combined as [? SUB]; eauto.
  Qed.
  
  Theorem typing_substitution_front : forall t' A G t b B,
       val v
    -> typing G t' v A
    -> typing ((t',A) :: G) t b B
    -> typing G t ({0 ~> v}b) ({0 ~> v} B).
  Proof.
    intros t' A G t b B ? ? HTyp;
    replace ((t',A) :: G) with (nil ++ (t',A) :: G) 
         in HTyp by (simpl; auto);
    eapply typing_substitution in HTyp; 
    simpl in *; autorewrite with step in *; eauto.
  Qed.

  Theorem typing_substitution_front_lifted : forall t' A G t b B,
       val v -> typing G t' v A
    -> typing ((t',A) :: G) t b (lift B 1 0)
    -> typing G t ({0 ~> v} b) B.
  Proof.
    intros t' A G t b B HVal HTypV HTyp;
    eapply typing_substitution_front in HTyp; autorewrite with step in *;
    eauto.
  Qed.
 
  Theorem typing_substitution_end_closed : forall G t' B t a A,
       val v -> fvs v = 0 -> typing nil t' v B
    -> typing (G ++ (t',B) :: nil) t a A
    -> typing (open_c 0 v G) t ({length G ~> v} a) ({length G ~> v}A).
  Proof.
    intros; match goal with
    | [ HVal : val ?v', HTyp : typing (_ ++ _ :: _) _ _ _ |- _ ] =>
      apply typing_substitution in HTyp; eauto;
      autorewrite with step in *; rewrite app_nil_r in HTyp; eauto
    end.
  Qed.

  Theorem typing_substitution_front_closed : forall t' A G t b B,
       val v
    -> fvs B = 0
    -> typing G t' v A
    -> typing ((t',A) :: G) t b B
    -> typing G t ({0 ~> v}b) B.
  Proof.
      intros t' A G t b B ? ? ? HTyp;
      eapply typing_substitution_front in HTyp; eauto;
      rewrite (open_closed_id B) in HTyp; eauto.
  Qed.
End Substitution.
Hint Resolve typing_substitution ctx_wf_substitution.
Hint Resolve typing_substitution_front.
Hint Immediate typing_substitution_front_lifted.
Hint Extern 2 =>
  match goal with
  | [   HT : typing ((?t'',?A')::?G) _ ?a exp_star
     |- typing ?G ?t ({0 ~> ?v'}?a) ?B ] =>
    let HClosed := fresh "HClosed" in
    assert (fvs B = 0) as HClosed by (simpl; eauto);
    apply typing_substitution_front_closed
     with (v := v') (t' := t'') (A := A'); simpl
  end.

Ltac trysubs :=
  repeat (match goal with
  | [ H : typing (_ :: (?t,?A') :: _) _ ?b _ 
     |- typing _ _ ({0 ~> ?v} (?P ?b)) ({0 ~> ?v}?B) ] =>
    apply typing_substitution_front with (t' := t) (A := A'); auto; eauto;
    try eassumption
  | [ H : typing ((?t,?A') :: _) _ ?b _ 
     |- typing _ _ ({0 ~> ?v} ?b) ({0 ~> ?v}?B) ] =>
    apply typing_substitution_front with (t' := t) (A := A'); auto; eauto;
    try eassumption
  | [ H : typing ((?t,?A') :: _) _ ?b (lift _ _ _)
     |- typing _ _ ({0 ~> ?v} ?b) ?B ] =>
    apply typing_substitution_front_lifted with (t' := t) (A := A'); eauto;
    try eassumption
  end).  

Ltac mycrush_subs := mycrush' trysubs.
