Require Import List.
Require Import Arith.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_reduction_basics.
Require Import step_typing_basics.
Require Import step_substitution.
Require Import step_preservation.

Require Import Program.Equality.
Require Import Omega.

Variable b v : exp.

Theorem TConv_n : forall n A G p a b1 b2 t,
     typing G th_log p (exp_eq b1 b2)
  -> typing G t a ({n ~> b1} A)
  -> typing G th_log ({n ~> b2} A) exp_star
  -> typing G t a ({n ~> b2} A).
Proof.
  intros n; intros; destruct (open_index_swap A n 0) as [? Heq];
  rewrite Heq in *; mycrush.
Qed.

Theorem mppreservation : forall G t a b A,
  typing G t a A -> mpstep a b -> typing G t b A.
Proof.
  induction 2; eauto using pstep_preservation.
Qed.

Hint Immediate mppreservation.


Theorem substitution_combined : 
     (   forall G t a A, typing G t a A
      -> forall G1 t' B G2, G = G1 ++ (t',B) :: G2
          -> mpstep b v
          -> val v
          -> typing G2 t' b B
          -> typing ((open_c 0 b G1) ++ G2) t
                    ({length G1 ~> lift b (length G1) 0} a)
                    ({length G1 ~> lift b (length G1) 0} A))
  /\ (   forall G, ctx_wf G
      -> forall G1 t' B G2, G = G1 ++ (t',B) :: G2
          -> mpstep b v
          -> val v
          -> typing G2 t' b B
          -> ctx_wf ((open_c 0 b G1) ++ G2)).
Proof.
  apply typing_combined_ind;
  intros; subst; try solve [match goal with
  | [ |- context[ exp_var (length _) ] ] =>
    (* TVar *)
    db_var_case; subst; simpl in *; invert_list_equalities; find_cons;
    inst_equalities; mycrush; eauto; simpl;
    match goal with
    | [ |- context[ ?n + S ?m ] ] =>
      replace (n + S m) with ((S n) + m) by omega
    end;
    [ rewrite open_lift_commute by omega | simpl; rewrite app_assoc];
    apply TVar; autorewrite with step; eauto
  | [ |- _ ] =>
    inst_equalities; mycrush; eauto 6;
    repeat (match goal with
    | [ |- context [{length ?n ~> _} ({0 ~> _} _)] ] =>
      rewrite open_open_swap with (n1 := length n) by omega
    | [ H : context [{length ?n ~> _} ({0 ~> _} _)] |- _ ] =>
      rewrite open_open_swap with (n1 := length n) in H by omega
    end); mycrush; eauto
  end].

  Ltac extract_arr := repeat (
  match goal with
  | [ H : context[ exp_arr ({?n ~> lift ?b ?n 0}?A) ({S ?n ~> lift ?b (S ?n) 0}?B)] |- _ ] =>
    replace (exp_arr ({n ~> lift b n 0}A) ({S n ~> lift b (S n) 0}B))
       with ({n ~> lift b n 0}(exp_arr A B)) in H by (simpl; autorewrite with step; eauto)
  end).

Theorem open_preserves_mpstep_l : forall b a a',
  mpstep a a' -> forall n, mpstep ({n ~> a}b) ({n ~> a'}b).
Proof.
  induction 1; eauto.
Qed.
Hint Immediate open_preserves_mpstep_l.

  inst_equalities.
  mycrush.  
  Check TEqArrInv2.
  assert (mpstep (lift b (length G1) 0) (lift v (length G1) 0)) by  eauto.
  assert (mpstep ({length G1 ~> lift b (length G1) 0}v0)
                 ({length G1 ~> lift v (length G1) 0}v0)) by  eauto.
  assert (H0orig := H0).
  apply mppreservation with (b := {length G1 ~> lift v (length G1) 0}v0) in H0; eauto.
  assert (val ({length G1 ~> lift v (length G1) 0} v0)) by eauto.
  extract_open_lift.
  assert (Foo := TEqArrInv2 _ _ _ _ _ _ _ _ _ H H0 H7).
    repeat (match goal with
    | [ |- context [{length ?n ~> _} ({0 ~> _} _)] ] =>
      rewrite open_open_swap with (n1 := length n) by omega
    | [ H : context [{length ?n ~> _} ({0 ~> _} _)] |- _ ] =>
      rewrite open_open_swap with (n1 := length n) in H by omega
    end); mycrush; eauto.   
  assert (H1orig := H1).
  apply mppreservation with (b := 
       (exp_eq
             (({0 ~> ({length G1 ~> lift v (length G1) 0}) v0})
                (({S (length G1) ~> lift b (S (length G1)) 0}) B1))
             (({0 ~> ({length G1 ~> lift v (length G1) 0}) v0})
                (({S (length G1) ~> lift b (S (length G1)) 0}) B2)))) in H1;
     [ | extract_open_lift; eauto].
  assert (Foo2 := Foo H1); clear Foo.
  extract_open_lift.
  apply TConv with (p := exp_join) (b1 := {length G1 ~> lift v (length G1) 0} v0).
  eapply TJoin; eauto.    eauto.  eauto.


  inst_equalities.  mycrush.
  assert (mpstep (lift b (length G1) 0) (lift v (length G1) 0)) by  eauto.
  assert (mpstep ({length G1 ~> lift b (length G1) 0}v0)
                 ({length G1 ~> lift v (length G1) 0}v0)) by  eauto.
  assert (H0orig := H0).
  apply mppreservation with (b := {length G1 ~> lift v (length G1) 0}v0) in H0; eauto.
  assert (val ({length G1 ~> lift v (length G1) 0} v0)) by eauto.
  extract_open_lift.
  assert (Foo := TEqSigmaInv2 _ _ _ _ _ _ _ _ _ H H0 H7).
    repeat (match goal with
    | [ |- context [{length ?n ~> _} ({0 ~> _} _)] ] =>
      rewrite open_open_swap with (n1 := length n) by omega
    | [ H : context [{length ?n ~> _} ({0 ~> _} _)] |- _ ] =>
      rewrite open_open_swap with (n1 := length n) in H by omega
    end); mycrush; eauto.   
  assert (H1orig := H1).
  apply mppreservation with (b := 
       (exp_eq
             (({0 ~> ({length G1 ~> lift v (length G1) 0}) v0})
                (({S (length G1) ~> lift b (S (length G1)) 0}) B1))
             (({0 ~> ({length G1 ~> lift v (length G1) 0}) v0})
                (({S (length G1) ~> lift b (S (length G1)) 0}) B2)))) in H1;
     [ | extract_open_lift; eauto].
  assert (Foo2 := Foo H1); clear Foo.
  extract_open_lift.
  apply TConv with (p := exp_join) (b1 := {length G1 ~> lift v (length G1) 0} v0).
  eapply TJoin; eauto.    eauto.  eauto.

  inst_equalities; mycrush.

  Focus 3.
  inst_equalities; mycrush.
  eapply TJoin with (c := {length G1 ~> lift v (length G1) 0}c).

Theorem MPTrans : forall a b c,
  mpstep a b -> mpstep b c -> mpstep a c.
Proof.
  induction 1; intros; simpl; eauto.
Qed.
   idtac.
   eapply MPTrans with (b := {length G1 ~> lift v (length G1) 0}a).
   eauto using open_preserves_mpstep_l.  eauto.

   eapply MPTrans with (b := {length G1 ~> lift v (length G1) 0}b0).
   eauto using open_preserves_mpstep_l.  eauto.

   eauto.  eauto.
   
  eauto.

    repeat (match goal with
    | [ |- context [{length ?n ~> _} ({0 ~> _} _)] ] =>
      rewrite open_open_swap with (n1 := length n) by omega
    | [ H : context [{length ?n ~> _} ({0 ~> _} _)] |- _ ] =>
      rewrite open_open_swap with (n1 := length n) in H by omega
    end); mycrush; eauto.  

    apply Foo.
    eapply TEqArrInv2.
extract_open_lift.


Theorem sub_val_goes_to_val : forall b vb v,
    


    eapply TEqArrInv2.  eassumption.  eauto.  eauto.
    
    eapply TEqArrInv2 in H with .
Qed.

