Require Import List.
Require Import Arith.
Require Import step_defs.
Require Import step_db_basics.
Require Import Program.Equality.
Require Import Omega.
Require Import NPeano.

Scheme typing_mut_ind := Induction for typing Sort Prop
  with ctx_wf_mut_ind := Induction for ctx_wf Sort Prop.

Combined Scheme typing_combined_ind 
           from typing_mut_ind, ctx_wf_mut_ind.

Hint Rewrite <- app_assoc : step.

(**********************************************)
(*** Generalized versions of rules that say L *)
(*
Theorem TStar_gen : forall G t,
  ctx_wf G -> typing G t (exp_sort star) (exp_sort box).
Proof.
  destruct t; intros; eauto using TSub.
Qed.
Hint Extern 1 =>
  match goal with
  | [ |- typing ?G ?t (exp_sort star) _ ] =>
    first [ apply TStar | apply TStar_gen ]
  | [ |- typing ?G ?t (exp_sort ?s) _ ] =>
    is_evar s; first [ apply TStar | apply TStar_gen ]
  end.
*)
Theorem TNat_gen : forall G t,
  ctx_wf G -> typing G t exp_nat (exp_star).
Proof.
  destruct t; intros; eauto using TSub.
Qed.
Hint Extern 1 =>
  match goal with
  | [ |- typing ?G _ exp_nat _ ] =>
    first [ apply TNat | apply TNat_gen ]
  end.
(*
Theorem TMu_gen : forall G t A,
     typing ((t,exp_star) :: G) th_log A (exp_star)
  -> typing G t' (exp_mu A) (exp_star).
Proof.
  intros G t t' A H; apply TMu in H; destruct t'; eauto.
Qed.
Hint Extern 1 =>
  match goal with
  | [ |- typing ?G _ (exp_mu _ _) _ ] =>
    first [ apply TMu | apply TMu_gen]
  end.
*)
Theorem TVar1 : forall t t' A B G,
     ctx_wf ((t',A) :: (t,B) :: G)
  -> typing ((t',A) :: (t,B) :: G) t (exp_var 1) (lift B 2 0).
Proof.
  intros; match goal with
  | [ |- typing (?b1 :: ?b2 :: ?G) _ _ _ ] =>
      replace (b1 :: b2 :: G) with ((b1 :: nil) ++ (b2 :: G))
           in * by auto
  end; eauto.
Qed.
Hint Resolve TVar1.


Inductive Boxed (A : exp) (t : th) : exp -> Prop :=
  | BOne : Boxed A t (exp_at A t)
  | BMany : forall B t',
      Boxed A t B -> Boxed A t (exp_at B t').
(*
Theorem boxed_star_typable : forall t B,
     Boxed (exp_sort star) t B
  -> forall G t', ctx_wf G -> typing G t' B (exp_sort box).
Proof.
  induction 1; intros; eauto.
Qed.
Hint Resolve boxed_star_typable.
*)
(******************************************************)
(********* Some tactics *******************************)

Ltac find_contra :=
  match goal with
  | [ H : typing _ _ _ (exp_eq ?A ?B) |- _ ] =>
    let h1 := fresh "h1" in let h2 := fresh "h2" in
    evar (h1:head); assert (head_form A h1) by (subst h1; econstructor);
    evar (h2:head); assert (head_form B h2) by (subst h2; econstructor);
    assert (h1 <> h2) by (subst h1 h2; intros HContra; inversion HContra);
    subst h1 h2
  | [ H : typing _ _ _ (exp_eq ?A ?B'), HBoxed : Boxed _ _ ?B' |- _ ] =>
    is_var B'; inversion HBoxed; subst;
    match goal with
    | [ H : typing _ _ _ (exp_eq A ?B) |- _ ] =>
      let h1 := fresh "h1" in let h2 := fresh "h2" in
      evar (h1:head); assert (head_form A h1) by (subst h1; econstructor);
      evar (h2:head); assert (head_form B h2) by (subst h2; econstructor);
      assert (h1 <> h2) by (subst h1 h2; intros HContra; inversion HContra);
      subst h1 h2
    end
  end.

Hint Extern 2 => 
  match goal with
(*  | [ H : ?ha <> ?hb, HH1 : head_form ?A ?ha, HH2 : head_form ?B ?hb, HFO : FO ?C,
      HEq : typing _ ?t ?p' (exp_eq ?A ?B),
      HRandom : typing ((?t1,?C) :: ?G) ?tr ?ar ?Ar |- typing ((?t2,?C) :: ?G) ?t' ?ar _ ] =>
      is_var t1; is_var t2; require_neq t1 t2; destruct t1; destruct t2; match goal with
      | [ H : ?ha <> ?hb, HH1 : head_form ?A ?ha, HH2 : head_form ?B ?hb,
          HEq : typing _ ?t ?p' (exp_eq ?A ?B) |- _ ] =>
        eapply (TContra _ _ _ (lift p' 1 0) (lift A 1 0) (lift B 1 0))
      end; try eassumption *)
  | [ H : ?ha <> ?hb, HH1 : head_form ?A ?ha, HH2 : head_form ?B ?hb,
      HEq : typing _ ?t ?p' (exp_eq ?A ?B),
      HRandom : typing ?G ?tr ?ar ?Ar |- typing ?G _ ?ar _ ] =>
      eapply (TContra _ _ _ _ _ _ _ _ _ _ _ HEq HH1 HH2 H HRandom);
      try eassumption
  | [ H : ?ha <> ?hb, HH1 : head_form ?A ?ha, HH2 : head_form ?B ?hb,
      HEq : typing _ ?t ?p' (exp_eq ?A ?B),
      HRandom : typing (?b :: ?G) ?tr ?ar ?Ar |- typing (?b :: ?G) ?t' ?ar _ ] =>
      eapply (TContra (b :: G) tr t' (lift p' 1 0) (lift A 1 0) (lift B 1 0));
      try eassumption
  | [ H : ?ha <> ?hb, HH1 : head_form ?A ?ha, HH2 : head_form ?B ?hb,
      HEq : typing _ ?t ?p' (exp_eq ?A ?B) |- typing ?G _ ?foo _ ] =>
      is_evar foo;
      eapply (TContra _ _ _ _ _ _ _ _ exp_zero _ _ HEq HH1 HH2 H)
  | [ H : ?ha <> ?hb, HH1 : head_form ?A ?ha, HH2 : head_form ?B ?hb,
      HEq : typing _ ?t ?p' (exp_eq ?A ?B) |- typing ?G _ ?foo _ ] =>
      eapply (TContra _ _ _ _ _ _ _ _ _ _ _ HEq HH1 HH2 H);
      try eassumption
  end.

Ltac tconstructor :=
  first [
    match goal with | [ |- typing _ _ (exp_var ?n) _ ] =>
      match goal with 
        | [ |- typing _ _ _ (exp_at _ _) ] => fail 2
        | [ |- context [(_,(exp_at _ _)) :: _] ] => eapply TUnboxVal
        | [ H : typing _ _ ?v (exp_at _ _), H' : val ?v |- _ ] => eapply TUnboxVal
      end
    end
    | try find_contra; match goal with
      | [ Ha : head_form ?a ?ha, Hb : head_form ?b ?hb, Hneq : ?ha <> ?hb
         |- typing _ _ _ _ ] =>
        eapply TContra
      end
    | eapply TVar | eapply TApp | eapply TLam | eapply TRec | eapply TNRec
    | eapply TSuc 
    | match goal with
      | [ |- typing _ th_log ?v (exp_at _ th_prog) ] =>
        assert (val v) by eauto; eapply TAtLV
      | [ |- typing _ th_log _ (exp_at _ _) ] => eapply TAtL
      | [ |- typing _ th_prog _ (exp_at _ _) ] => eapply TAtP
      end
    | eapply TRoll | eapply TUnroll
    | eapply CWFNil | eapply CWFTyp | eapply CWFTrm
    | eapply TNat | eapply TArr | eapply TAt
    | eapply TMu | eapply TEq (* | eapply TStar *)
    | eapply TSigma | eapply TPair
    | match goal with
      | [ |- typing _ _ (exp_ncase _ _ _) _ ] => eapply TNCase
      | [ |- typing _ _ (exp_scase _ _ _) _ ] => eapply TSCase
      | [ |- typing _ _ (exp_pcase _ _) _   ] => eapply TPCase
      | [ |- typing _ _ ?v ?A ] =>
        match A with
        | exp_eq _ _ => fail 2
        | _ => idtac
        end; assert (val v) by eauto; assert (FO A) by eauto; eapply TFOVal
      | [ H : context [ typing _ th_log ?a ?A ] 
         |- typing ?G th_prog ?a ?A ] => eapply TSub
      end
  ].

Ltac inst_equalities := 
  repeat (match goal with
  | [ H : forall _ _, _ ++ _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ ++ _ :: _ ++ _ = _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, (_ ++ _ :: _) ++ _ :: _ = _ -> _ |- _ ] =>
    try rewrite <- app_assoc in H; try rewrite <- app_comm_cons in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H

  | [ H : forall _ _ _ _, _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall G, ctx_wf (G ++ ?G3) -> _,
      H2 : ctx_wf (?G2 ++ ?G3)
     |- context[lift_c _ (length ?G2) _] ] => let Hwf := fresh "Hwf" in 
    assert (Hwf := H _ H2); clear H; rename Hwf into H 
  | [ H : forall _ _ _ _, _ ++ _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, (_ ++ _ :: _) ++ _ :: _ = _ -> _ |- _ ] =>
    try rewrite <- app_assoc in H; try rewrite <- app_comm_cons in H;
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _ _, _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ _ eq_refl); clear H; rename H' into H
  | [ H : forall _ _, _ :: _ :: _ :: _ ++ _ = _ ++ _ -> _ |- _ ] =>
    let H' := fresh H "'" in
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    rewrite app_comm_cons in H;
    assert (H' := H _ _ eq_refl); clear H; rename H' into H

  | [ H : forall _ _ _ _ _ _ _, _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ ++ _ :: ?G = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing ?G _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H

  | [ H : forall _ _ _ _ _ _ _, _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ :: _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ ++ _ :: _ ++ (_,({_ ~> _}_)) :: _
                              = _ ++ (_,({_ ~> _}_)) :: _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    rewrite app_comm_cons in H; rewrite app_assoc in H;
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, (_ ++ (_,({_ ~> _}_)) :: _) ++ _
                              = _ ++ (_,({_ ~> _}_)) :: _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    rewrite <- app_assoc in H; rewrite <- app_comm_cons in H;
    let H' := fresh H "'" in
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  | [ H : forall _ _ _ _ _ _ _, _ :: _ :: _ :: _ ++ _ = _
             -> typing _ _ _ (exp_eq _ _) -> _,
      HE : typing _ _ _ (exp_eq _ _) |- _ ] =>
    let H' := fresh H "'" in
    repeat (rewrite app_comm_cons in H);
    assert (H' := H _ _ _ _ _ _ _ eq_refl HE); clear H; rename H' into H
  end).

Ltac mycrush' adjust :=
  try adjust; eauto;
  simpl in *; autorewrite with step in *;
  simpl in *; autorewrite with step in *; eauto;
  try (match goal with
  | [ HEq :  nil = ?G1 ++ _ |- ctx_wf _ ] =>
    destruct G1; inversion HEq; simpl in *; subst
  | [ HEq :  _ :: _ = ?G1 ++ _ |- ctx_wf _ ] =>
    destruct G1; inversion HEq; simpl in *; subst
  end);
  repeat (match goal with
    | [   H1 : ?A, H2 : ?A -> ?B |- _ ] =>
      let H2' := fresh H2 in
      assert (H2' := H2 H1); clear H2; rename H2' into H2
  end);
  try tconstructor;
  try (match goal with 
  | [ H : typing _ _ _ (exp_eq ?a ?b) |- typing _ _ _ ({0 ~> ?b} _) ] =>
    eapply TConv with (b1 := a); eauto
  | [ H : typing _ _ _ (exp_eq ?a ?b) |- typing _ _ _ ({0 ~> ?a} _) ] =>
    eapply TConv with (b1 := b); eauto
  | [ H : pstep ?a ?b |- typing _ _ _ ({0 ~> ?b} _) ] =>
    eapply TConv with (b1 := a)
  | [ H : pstep ?a ?b |- typing _ _ _ ({0 ~> ?a} _) ] =>
    eapply TConv with (b1 := b)
  | [ H : typing _ _ _ (exp_eq ?a ?b) |- typing _ _ _ ({0 ~> lift ?b ?n ?m} _) ] =>
    eapply TConv with (b1 := lift a n m); eauto
  | [ H : typing _ _ _ (exp_eq ?a ?b) |- typing _ _ _ ({0 ~> lift ?a ?n ?m} _) ] =>
    eapply TConv with (b1 := lift b n m); eauto
  | [ H : pstep ?a ?b |- typing _ _ _ ({0 ~> lift ?b ?n ?m} _) ] =>
    eapply TConv with (b1 := lift a n m)
  | [ H : pstep ?a ?b |- typing _ _ _ ({0 ~> lift ?a ?n ?m} _) ] =>
    eapply TConv with (b1 := lift b n m)
  end); try adjust;
  try (match goal with
  | [ |- typing _ th_log exp_join (exp_eq _ _) ] => idtac
  | [ |- typing _ th_prog exp_join (exp_eq _ _) ] => apply TSub
  | [ |- typing _ ?t exp_join (exp_eq _ _) ] => destruct t;
    try match goal with | [ |- typing _ th_prog _ _ ] => apply TSub end
  end);
  try (match goal with
  | [ H : pstep ?a ?b |- typing ?G th_log _ (exp_eq ?a ?b) ] =>
    eapply TJoin with (c := b); eauto
  | [ H : pstep ?a ?b |- typing ?G th_log _ (exp_eq ?b ?a) ] =>
    eapply TJoin with (c := b); eauto
  | [ H : pstep ?a ?b 
     |- typing ?G th_log _ (exp_eq (lift ?a ?n ?m) (lift ?b ?n ?m)) ] =>
    eapply TJoin with (c := lift b n m); eauto
  | [ H : pstep ?a ?b 
     |- typing ?G th_log _ (exp_eq (lift ?b ?n ?m) (lift ?a ?n ?m)) ] =>
    eapply TJoin with (c := lift b n m); eauto
  | [ H1 : pstep ?b1 ?a1, H2 : pstep ?b2 ?a2
     |- typing ?G th_log _ (exp_eq (?P ?a1 ?a2) (?P ?b1 ?b2)) ] =>
    eapply TJoin with (c := P a1 a2); eauto
  | [ H1 : pstep ?a1 ?b1, H2 : pstep ?a2 ?b2
     |- typing ?G th_log _ (exp_eq (?P ?a1 ?a2) (?P ?b1 ?b2)) ] =>
    eapply TJoin with (c := P b1 b2); eauto
  | [ H : pstep ?a ?b |- typing ?G th_log _ (exp_eq (?P ?a) (?P ?b)) ] =>
    eapply TJoin with (c := P b); eauto
  | [ H : pstep ?a ?b |- typing ?G th_log _ (exp_eq (?P ?b) (?P ?a)) ] =>
    eapply TJoin with (c := P b); eauto
  | [ H1 : pstep ?a ?c', H2 : pstep ?b ?c'
     |- typing _ th_log exp_join (exp_eq ?a ?b) ] => eapply TJoin with (c := c')
  end); eauto; try adjust;
  try tconstructor;
  try solve [simpl in *; autorewrite with step in *; simpl in *; eauto].

Ltac mycrush := mycrush' idtac.

(* These two theorems invert some equalities on lists, given the types of
   information we get out of the clarify_compares operation *)
Theorem app_eq_lengths_eq : forall {A} (G1 G2 G3 G4 : list A),
     G1 ++ G2 = G3 ++ G4
  -> length G1 = length G3
  -> G1 = G3 /\ G2 = G4.
Proof.
  induction G1; intros G2 G3 G4 Heq Hlen;
  destruct G3; simpl in *; inversion Hlen; auto;
  inversion Heq; subst; eauto.
  match goal with
  | [ IH : forall _ _ _, ?G1 ++ _ = _ -> length ?G1 = _ -> _ /\ _,
      Heq : ?G1 ++ _ = _,
      Hlen : length ?G1 = _ |- _ ] =>
      destruct (IH _ _ _ Heq Hlen); subst; eauto
  end.
Qed.

Theorem app_eq_lengths_lt : forall {A} (G1 G2 G3 G4 : list A),
     G1 ++ G2 = G3 ++ G4
  -> length G1 < length G3
  -> exists G5, G3 = G1 ++ G5 /\ G2 = G5 ++ G4.
Proof.
  induction G1; intros G2 G3 G4 Heq Hlen;
  destruct G3; simpl in *; subst; eauto;
  try solve [inversion Hlen];
  inversion Heq; subst;
  match goal with
  | [ IH : forall _ _ _, _ ++ _ = _ ++ _ -> _ -> exists _ , _,
      Heq : _ ++ _ = _ ++ _ |- _ ] =>
    destruct (IH _ _ _ Heq) as [G5 [Heq1 Heq2]]
  end; try omega; subst; eauto.
Qed.

(* This tactic just tries to apply the previous two theorems. *)
Ltac invert_list_equalities :=
  (* Begin by making the > case look like the < case. *)
  try match goal with
  | [ H1 : length ?G2 > length ?G1, H2 : ?G2 ++ _ = ?G1 ++ _ |- _ ] =>
    unfold gt in H1; apply eq_sym in H2
  end;
  match goal with
  | [ Heq : ?G1 ++ _ = ?G2 ++ _, Hlen : length ?G1 = length ?G2 |- _ ] =>
    destruct (app_eq_lengths_eq _ _ _ _ Heq Hlen); subst; clear Hlen Heq
  | [ Heq : ?G1 ++ _ = ?G2 ++ _, Hlen : length ?G1 < length ?G2 |- _ ] =>
    let G5 := fresh "G5" in
    let HEq1 := fresh "HEq1" in let HEq2 := fresh "HEq2" in
    destruct (app_eq_lengths_lt _ _ _ _ Heq Hlen) as [G5 [Heq1 Heq2]]; subst
  end.

(* this one looks for the situation where we know:

     length G1 < length (G1 ++ G5)

   Then G5 must be a cons, so we expose it *)
Ltac find_cons := 
  match goal with
  | [ Hlen : length ?G1 < length (?G1 ++ ?G5) |- _ ] =>
    destruct G5; 
      [  subst; rewrite app_length in Hlen; rewrite plus_comm in Hlen; 
         simpl in Hlen; destruct (lt_irrefl _ Hlen) 
       | subst; clear Hlen ];
    match goal with
    | [ Heq2 : _ :: _ = (_ :: _) ++ _ |- _ ] =>
      inversion Heq2; subst; clear Heq2
    end
  | [ HConsEq : _ :: _ = _ :: _ |- _ ] =>
    inversion HConsEq; subst; clear HConsEq
  end.


(*****************************************************************)
(* Some basic facts about ctx_wf and typ_wf that don't even need
   weakening, mixed with some Ltac to apply them easily. *)
Theorem ctx_wf_inversion_app : forall G1 G2,
  ctx_wf (G1 ++ G2) -> ctx_wf G2.
Proof.
  induction G1; intros; simpl in *; auto; match goal with
  | [ H : ctx_wf _ |- _ ] => inversion H
  end; subst; auto.
Qed.
Hint Immediate ctx_wf_inversion_app.

Ltac notProvenYet P :=
  match goal with
  | [ _ : P |- _ ] => fail 1
  | _ => idtac
  end.

Ltac invert_ctx_wf := simpl in *;
  try match goal with
  | [ H : ctx_wf (_ ++ ?G) |- _ ] =>
    notProvenYet (ctx_wf G); let HWF := fresh "HWF" in
    assert (HWF := ctx_wf_inversion_app _ _ H); invert_ctx_wf
  | [ H : ctx_wf ((_,?A) :: ?G) |- _ ] =>
     notProvenYet(ctx_wf G); inversion H; subst; invert_ctx_wf
  | _ => idtac
  end.

Theorem typing_regularity_ctx : forall G t a A,
  typing G t a A -> ctx_wf G.
Proof.
  induction 1;  invert_ctx_wf; eauto.
Qed.
Hint Immediate typing_regularity_ctx.

(**** Free variables ***)
Hint Resolve Nat.max_lub.
Hint Resolve Nat.max_lub_l.
Hint Resolve Nat.max_lub_r.

Theorem typing_fvs : 
  forall G t a A, typing G t a A
    -> fvs a <= length G.
Proof.
  induction 1; intros; auto; subst; simpl in *;
  autorewrite with step in *; simpl in *; eauto; omega.
Qed.

Theorem typing_fvs_closed_exp : forall t a A,
  typing nil t a A -> fvs a = 0.
Proof.  
  intros t a A H; apply typing_fvs in H; simpl in *; omega.
Qed.
Hint Immediate typing_fvs_closed_exp.


(*************************************************)  
(****************** Weakening ********************)
Theorem weakening_combined :
     (   forall G t a A, typing G t a A 
      -> forall G1 G3, G = G1 ++ G3
      -> forall G2, let n := length G2 in let m := length G1 in
              ctx_wf (G2 ++ G3)
           -> typing (lift_c G1 n 0 ++ G2 ++ G3) t
                     (lift a n m) (lift A n m))
  /\ (   forall G, ctx_wf G
      -> forall G1 G3, G = G1 ++ G3
      -> forall G2, ctx_wf (G2 ++ G3)-> 
           let n := length G2 in
           ctx_wf ((lift_c G1 n 0) ++ G2 ++ G3)).
Proof.
  apply typing_combined_ind;
  intros; subst; eauto;
  repeat (match goal with | [n := length _ : nat |- _] => subst n end);
  simpl in *;
  try match goal with
  (* the ctx_wf cases *)
  | [ HEq : _ = ?G1 ++ _ |- ctx_wf _ ] =>
    destruct G1; inversion HEq; simpl in *; inst_equalities; eauto
  (* The var case *)
  | [ |- typing _ _ (exp_var _) _ ] =>
    clarify_compares; subst; eauto; invert_list_equalities;
    [ autorewrite with step; rewrite app_assoc;
      apply TVar; autorewrite with step; eauto
  
    | find_cons;  autorewrite with step in *; simpl; eauto;
      inst_equalities; match goal with 
      | [ |- context [ ?n + S ?m ] ] =>
        replace (n + S m) with ((S n) + m) by omega
      end; autorewrite with step in *; simpl in *; eauto
  
    | autorewrite with step; repeat (rewrite app_assoc); apply TVar;
      autorewrite with step; try omega; eauto ]
  end; try solve [inst_equalities; mycrush; simpl; extract_open_lift; eauto].
Qed.

Theorem typing_weakening : forall G1 G2 G3 t a A,
     typing (G1 ++ G3) t a A 
  -> ctx_wf (G2 ++ G3)
  -> typing (lift_c G1 (length G2) 0 ++ G2 ++ G3) t 
            (lift a (length G2) (length G1)) 
            (lift A (length G2) (length G1)).
Proof.
  intros; destruct weakening_combined as [ Htyp _ ]; eauto.
Qed.

Theorem ctx_wf_weakening : forall G1 G2 G3,
     ctx_wf (G1 ++ G3)
  -> ctx_wf (G2 ++ G3)
  -> ctx_wf (lift_c G1 (length G2) 0 ++ G2 ++ G3).
Proof.
  intros; destruct weakening_combined as [ _ Hctx ]; eauto.
Qed.

(***********************************************************)
(******* Automation-friendly versions of weakening *********)
Theorem typing_weakening_pre_app : forall G1 G2 t a A,
     typing G2 t a A
  -> ctx_wf (G1 ++ G2)
  -> typing (G1 ++ G2) t (lift a (length G1) 0)
            (lift A (length G1) 0).
Proof.
  intros;
  apply typing_weakening with (G1 := nil) (G2 := G1) in H; eauto.
Qed.
Hint Resolve typing_weakening_pre_app.

Hint Extern 2 =>
  match goal with
  | [ |- typing (?G1 ++ _) _ (lift _ ?n 0) (lift _ ?n 0) ] =>
    replace n with (length G1) by (simpl; autorewrite with step; auto)
  end.

Theorem typing_weakening_cons : forall G t' B t a A,
     typing G t a A
  -> typing G t' B exp_star
  -> typing ((t',B) :: G) t (lift a 1 0) (lift A 1 0).
Proof.
  intros G t' B t a A HTyping HTyp_wf;
  apply typing_weakening_pre_app with (G1 := (t',B) :: nil) in HTyping;
  simpl in *; eauto.
Qed.
Hint Resolve typing_weakening_cons.

Theorem typing_weakening_cons_cons : forall G t'1 B1 t'2 B2 t a A,
     typing G t a A
  -> typing G t'1 B1 exp_star
  -> typing ((t'1,B1)::G) t'2 B2 exp_star
  -> typing ((t'2,B2)::(t'1,B1)::G) t (lift a 2 0) (lift A 2 0).
Proof.
  intros; match goal with
  | [ H : typing _ _ _ ?A |- typing (?b1 :: ?b2 :: ?G) _ _ (lift ?A _ _) ] =>
    apply typing_weakening_pre_app with (G1 := b1 :: b2 :: nil) in H
  end; simpl in *; eauto.
Qed.
Hint Resolve typing_weakening_cons_cons.

Theorem typing_weakening_cons_star : forall G t' t a A,
     typing G t a A
  -> typing ((t',exp_star) :: G) t (lift a 1 0) (lift A 1 0).
Proof.
  intros G t' t a A HTyping;
  apply typing_weakening_pre_app with (G1 := (t',exp_star) :: nil) in HTyping;
  simpl in *; eauto.
Qed.
Hint Resolve typing_weakening_cons_star.

Theorem star_has_no_type' : forall G t a A,
     typing G t a A
  -> a = exp_star
  -> False.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_star |- _] => inversion H; subst; clear H
  end; eauto.
Qed.  

Theorem star_has_no_type : forall G t A,
   typing G t exp_star A -> False.
Proof.
  intros; eauto using star_has_no_type'.
Qed.


Theorem typing_weakening_app_cons : forall G1 t' B G2 t a A,
     typing G2 t a A
  -> ctx_wf (G1 ++ (t',B) :: G2)
  -> typing (G1 ++ (t',B) :: G2) t (lift a (S (length G1)) 0)
            (lift A (S (length G1)) 0).
Proof.
  intros; invert_ctx_wf;
  match goal with
  | [ H : typing _ _ exp_star _ |- _ ] =>
    apply star_has_no_type in H; inversion H
  | [ Htyp : typing ?G2 _ ?a _ |- typing (?G1' ++ (?t,?B') :: ?G2) _ (lift ?a _ _) _ ] =>
    match B' with
    | exp_star => eapply typing_weakening_cons_star with (t' := t) in Htyp
    | _ => eapply typing_weakening_cons with (t' := t) (B := B') in Htyp
    end; eauto;
    apply typing_weakening_pre_app with (G1 := G1') in Htyp;
    autorewrite with step in *; eauto
  end.
Qed.
Hint Resolve typing_weakening_app_cons.

Theorem typing_weakening_app_cons_closed_typ : forall G1 t' B G2 t a A,
     typing G2 t a A
  -> fvs A = 0
  -> ctx_wf (G1 ++ (t',B) :: G2)
  -> typing (G1 ++ (t',B) :: G2) t (lift a (S (length G1)) 0) A.
Proof.
  intros G1 t' B G2 t a A Htyp HFV HWf;
  replace A with (lift A (S (length G1)) 0) by (autorewrite with step; auto);
  eauto.
Qed.
Hint Resolve typing_weakening_app_cons_closed_typ.
  
(* these three just unfold lift_t, which is getting in the way of automation *)
Theorem typing_weakening_cons_eq : forall G t' B p a b,
     typing G th_log p (exp_eq a b)
  -> typing G t' B exp_star
  -> typing ((t',B) :: G) th_log (lift p 1 0) 
            (exp_eq (lift a 1 0) (lift b 1 0)).
Proof.
  intros ? ? ? ? ? ? Htyp ?; eapply typing_weakening_cons in Htyp; eauto.
Qed.

Theorem typing_weakening_app_cons_eq : forall G1 G2 t' B p a b,
     typing G2 th_log p (exp_eq a b)
  -> ctx_wf (G1 ++ (t',B) :: G2)
  -> typing (G1 ++ (t',B) :: G2) th_log (lift p (S (length G1)) 0)
         (exp_eq (lift a (S (length G1)) 0) (lift b (S (length G1)) 0)).
Proof.
  intros ? ? ? ? ? ? ? HTyp ?; eapply typing_weakening_app_cons in HTyp;
  simpl in *; eauto.
Qed.

Theorem typing_weakening_cons_cons_eq : forall G tc C tb B p a b,
     typing G th_log p (exp_eq a b)
  -> typing G tb B exp_star
  -> typing ((tb,B) :: G) tc C exp_star
  -> typing ((tc,C) :: (tb,B) :: G) th_log (lift p 2 0) 
            (exp_eq (lift a 2 0) (lift b 2 0)).
Proof.
  intros; match goal with
  | [ H : typing _ _ _ (exp_eq _ _) |- typing (?b1 :: ?b2 :: ?G) _ _ _ ] =>
    apply typing_weakening_pre_app with (G1 := b1 :: b2 :: nil) in H
  end; simpl in *; eauto.
Qed.
Hint Resolve typing_weakening_cons_eq.
Hint Resolve typing_weakening_app_cons_eq.
Hint Resolve typing_weakening_cons_cons_eq.

Theorem typing_weakening_cons_typ : forall G t' B t a,
     typing G t a exp_star
  -> typing G t' B exp_star
  -> typing ((t',B)::G) t (lift a 1 0) exp_star.
Proof.
  intros; replace exp_star with (lift exp_star 1 0) by auto; eauto. 
Qed.
Hint Resolve typing_weakening_cons_typ.

Theorem typing_weakening_cons_cons_typ : forall G t'1 B1 t'2 B2 t a,
     typing G t a exp_star
  -> typing G t'1 B1 exp_star
  -> typing ((t'1,B1)::G) t'2 B2 exp_star
  -> typing ((t'2,B2)::(t'1,B1)::G) t (lift a 2 0) exp_star.
Proof.
  intros; match goal with
  | [ H : typing _ _ ?a _ |- typing (?b1 :: ?b2 :: ?G) _ (lift ?a _ _) _ ] =>
    apply typing_weakening_pre_app with (G1 := b1 :: b2 :: nil) in H
  end; simpl in *; eauto.
Qed.
Hint Resolve typing_weakening_cons_cons_typ.

(* A little lemma to help proving things at th_prog, when we know they check
   at _some_ th *)
Theorem typing_promotion_gen : forall G t a A,
  typing G t a A -> typing G th_prog a A.
Proof.  intros; destruct t; eauto.  Qed.
Hint Immediate typing_promotion_gen.

Theorem typing_subsumption_gen : forall G t a A,
  typing G th_log a A -> typing G t a A.
Proof.  intros; destruct t; eauto.  Qed.
Hint Immediate typing_subsumption_gen.


Hint Extern 3 =>
  match goal with
  | [ |- typing (?G ++ _) _ _ ({_ ~> lift _ _ _} (lift _ (length ?G) _)) ] =>
    rewrite <- lift_open_commute by eauto; apply typing_weakening_pre_app 
  | [ |- typing (?G ++ (_,({_ ~> ?b}_)) :: _) _ (exp_var _) 
            ({_ ~> lift ?b _ _} (lift _ (S (length ?G)) _)) ] =>
    rewrite <- lift_open_commute by eauto
  end.

(* A general version of TSub *)
Theorem TSubGen : forall G t a A,
  typing G t a A -> typing G th_prog a A.
Proof.
  destruct t; intros; eauto using TSub.
Qed.  
Hint Immediate TSubGen.
