Require Import Arith.
Require Import Omega.
Require Import Bool.
Require Import List.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_reduction_basics.
Require Import step_substitution.
Require Import step_preservation.
Require Import step_interp.
Require Import step_interp_facts.

Hint Extern 2 =>
  match goal with
  | [ |- _ <= _ ] => omega
  end.

Theorem TICongLam : forall rho b n,
    TermInterp rho (exp_lam b) n
  = exp_lam (TermInterp rho b (S n)).
Proof.
  induction rho; simpl; intros; eauto;
  autorewrite with steps; eauto.
Qed.

Theorem TICongApp : forall rho a b n,
    TermInterp rho (exp_app a b) n
  = exp_app (TermInterp rho a n) (TermInterp rho b n).
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem TICongFun : forall rho a n,
    TermInterp rho (exp_fun a) n 
  = exp_fun (TermInterp rho a (S (S n))).
Proof.
  induction rho; intros; simpl; eauto;
  autorewrite with steps; rewrite plus_comm; eauto.
Qed.

Theorem TICongZero : forall rho n,
    TermInterp rho exp_zero n = exp_zero.
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem TICongSucc : forall rho a n,
    TermInterp rho (exp_succ a) n = exp_succ (TermInterp rho a n).
Proof. induction rho; intros; simpl; eauto. Qed.

Theorem TICongNatCase : forall rho a b1 b2 n,
   TermInterp rho (exp_natcase a b1 b2) n
 = exp_natcase (TermInterp rho a n)
            (TermInterp rho b1 n) (TermInterp rho b2 (S n)).
Proof. induction rho; intros; simpl; autorewrite with steps; eauto. Qed.

Theorem TICongSumCase : forall rho a b1 b2 n,
   TermInterp rho (exp_sumcase a b1 b2) n
 = exp_sumcase (TermInterp rho a n)
            (TermInterp rho b1 (S n)) (TermInterp rho b2 (S n)).
Proof. induction rho; intros; simpl; autorewrite with steps; eauto. Qed.

Theorem TICongRoll : forall rho a n,
   TermInterp rho (exp_roll a) n = exp_roll (TermInterp rho a n).
Proof.
  induction rho; intros; simpl; eauto.
Qed.

Theorem TICongUnroll : forall rho a n,
   TermInterp rho (exp_unroll a) n = exp_unroll (TermInterp rho a n).
Proof.
  induction rho; intros; simpl; eauto.
Qed.

Theorem TICongInl : forall rho a n,
  TermInterp rho (exp_inl a) n = exp_inl (TermInterp rho a n).
Proof.
  induction rho; intros; simpl; eauto.
Qed.

Theorem TICongInr : forall rho a n,
  TermInterp rho (exp_inr a) n = exp_inr (TermInterp rho a n).
Proof.
  induction rho; intros; simpl; eauto.
Qed.


Hint Rewrite TICongLam TICongFun TICongApp
             TICongZero TICongSucc TICongNatCase TICongSumCase
             TICongRoll TICongUnroll TICongInr TICongInl: steps.


Ltac ASCongSolver := 
  repeat (instantiate; match goal with
  | [ H : astep ?a ?b |- astep ?a ?b ] => apply H

(*  | [ HASBox : astep (exp_box ?a) (exp_box ?b) |- _ ] =>
    let HASBox' := fresh HASBox in
    assert (astep a b) as HASBox' by (eauto using ASBoxInversion);
    clear HASBox; rename HASBox' into HASBox *)

  | [ HAS : astep ?a ?a' |- astep (exp_app ?a ?c) _ ] =>
    apply ASTrans with (b := exp_app a' c);
    [ eauto using ASApp1 | ]
  | [ HAS : astep ?c ?c' |- astep (exp_app ?a ?c) _ ] =>
    apply ASTrans with (b := exp_app a c');
    [ eauto using ASApp2 | ]

(*  | [ HAS : astep ?a ?a' |- astep (exp_unbox ?a ?c) _ ] =>
    apply ASTrans with (b := exp_unbox a' c);
    [ eauto using ASUnbox1 | ] *)

  | [ HAS : astep ?a ?a' |- astep (exp_natcase ?a ?b1 ?b2) _ ] =>
    apply ASTrans with (b := exp_natcase a' b1 b2);
    [ eauto using ASNatCase1 | ]

  | [ HAS : astep ?a ?a' |- astep (exp_sumcase ?a ?b1 ?b2) _ ] =>
    apply ASTrans with (b := exp_sumcase a' b1 b2);
    [ eauto using ASSumCase1 | ]

  | [ |- astep ({?n ~> ?v} (TermInterp _ _ (S ?n))) _ ] =>
    erewrite TIOpenSwap; [ | solve [eauto] ];
    simpl; autorewrite with steps

  | [ |- astep ?a _ ] =>
    try (eapply ASTrans; [ solve [ apply ASStep; eauto ]; fail 2 | ])
  end).

Ltac mstep_inversion :=
  repeat (match goal with
(*  | [ HMS : mstep _ (exp_box _) _ |- _ ] =>
    apply MSBoxInversion in HMS; destruct HMS as [? [? HMS]] *)
  | [ HMS : mstep _ (exp_inl _) _ |- _ ] =>
    apply MSInlInversion in HMS; destruct HMS as [? [? HMS]]
  | [ HMS : mstep _ (exp_inr _) _ |- _ ] =>
    apply MSInrInversion in HMS; destruct HMS as [? [? HMS]]
  | [ HMS : mstep _ (exp_succ _) _ |- _ ] =>
    apply MSSuccInversion in HMS; destruct HMS as [? [? HMS]]
  | [ HMS : mstep _ (exp_roll _) _ |- _ ] =>
    apply MSRollInversion in HMS; destruct HMS as [? [? HMS]]
  | [ HV : val ?c, HMS : mstep _ (exp_unroll _) ?c |- _ ] =>
    apply mstep_unroll_val_inversion in HMS; auto;
    destruct HMS as [? [? HMS]]; subst
  end; subst).

Ltac mstep_val_inversion :=
  repeat (match goal with
  | [ HAS : astep _ _ |- _ ] => inversion HAS; subst; clear HAS
  end);
  mstep_inversion; val_inversion;
  try (match goal with
  | [ HMS1 : mstep _ ?a ?b, HMS2 : mstep _ ?a ?b' |- _ ] =>
    let HV1 := fresh "HV1" in let HV2 := fresh "HV2" in 
      assert (val b) as HV1 by eauto;
      assert (val b') as HV2 by eauto;
      destruct (mstep_val_deterministic _ _ _ _ _ HMS1 HMS2 HV1 HV2); 
      subst; clear HV1; clear HV2 
  end).

Theorem wfTermEnv_downward_closure_leq : forall k j G rho,
     wfTermEnv k G rho
  -> j <= k
  -> wfTermEnv j G rho.
Proof.
  intros; match goal with
  | [ Hle : _ <= _ |- _ ] =>
    apply le_lt_or_eq in Hle; destruct Hle;
    subst; eauto
  end.
Qed.
Hint Immediate wfTermEnv_downward_closure_leq.

Theorem wfTermEnv_downward_closure_minus : forall k j G rho,
     wfTermEnv k G rho
  -> wfTermEnv (k - j) G rho.
Proof.
  intros; assert (k - j <= k) by omega; eauto.
Qed.
Hint Immediate wfTermEnv_downward_closure_minus.

Hint Extern 4 =>
  match goal with
  | [ H : interp (?k, ?A, ?cv) ?t ?a |- interp (?j, ?A, ?cv) ?t ?a ] =>
    assert (j < k) by omega; apply (downward_closure k)
  end.


Theorem interp_soundness : forall k G t a A,
     typing G t a A 
  -> forall rho, wfTermEnv k G rho 
              -> interp (k,A,Comp) t (TermInterp rho a 0).
Proof.
  induction k as [k IHWf] using lt_wf_ind;
  induction 1; intros; subst; eauto; autorewrite with steps;
  repeat (match goal with
  | [ Hwf : wfTermEnv _ ?G _,
      IH : forall _, wfTermEnv _ ?G _ -> interp _ _ _ |- _ ] =>
    let IH' := fresh IH in assert (IH' := IH _ Hwf); clear IH;
    rename IH' into IH
  end);
  try match goal with
  | [ |- interp _ _ (TermInterp _ (exp_var _) _) ] =>
    (* var case *)
    eauto using comp_contains_values
  | [ H : interp _ th_log ?a |- interp _ th_prog ?a ] =>
    (* subsumption case *)
    eauto
  end; try match goal with
  (* app case *)
  | [ |- interp _ _ (exp_app _ _) ] => 
    unfold_interp; unfold_interp_hyps; intros; autorewrite with steps in *;
    match goal with
    | [ HAS1 : astep (TermInterp ?rho ?a _) (exp_lam ?va),
        HAS2 : astep (TermInterp _ ?b' _) ?vb,
        IHa : forall _ _, _ -> interp (_,?A,?Val) _ _
                            -> interp _ _ ({_ ~> _} ?va),
        IHb : interp (?k,?A,?Val) _ ?vb
       |- exists _, astep (exp_app (TermInterp _ ?a _) (TermInterp _ ?b' _)) _
                 /\ _ ] =>
      assert (interp (k,B,Comp) th_log ({0 ~> vb}va)) by eauto;
      unfold_interp_hyps; exists v2;
      split; try ASCongSolver; eauto
    | [ IHb : forall _ _, _ -> mstep _ (TermInterp _ ?b _) _
                            -> _ -> interp _ _ _,
        IHa : forall _ _, _ -> mstep _ (TermInterp _ ?a _) _
                            -> _ -> interp _ _ _,
        HMS : mstep _ (exp_app (TermInterp _ ?a _) (TermInterp _ ?b _)) _
       |- _ ] =>
      apply mstep_app_val_inversion in HMS; auto;
      destruct HMS as [j1 [j2 [j3 [a' [vb [HVvb [HMath [HMSb 
                       [ [HMSa HMSbeta] | [HMSa HMSbeta] ]]]]]]]]];
      apply IHa in HMSa; auto;
      apply IHb in HMSb; auto;
      unfold_interp_hyps;
      apply downward_closure with (j := k - (1 + j1 + j2)) in HMSb; auto;
      apply HMSa in HMSb; auto; unfold_interp_hyps;
      typing_inversion;
      try (erewrite lift_closed_id_gen in HMSbeta); eauto;
      apply HMSb in HMSbeta; auto;
      match goal with
      | [ HI : interp (?k1,?B,Val) th_prog ?v
         |-    interp (?k2,?B,Val) th_prog ?v ] =>
        replace k2 with k1 by omega; auto
      end
    end
  (* fun case *)  (* CHECKED *)
  | [ |- interp (_,typ_arr ?A ?B, Comp) th_prog
                (exp_fun (TermInterp ?rho ?a _)) ] =>
    unfold_interp; intros; step_val_elim;
    let HTyp1 := fresh "HTyp1" in let HTyp2 := fresh "HTyp2" in
    let IHInterp := fresh "IHInterp" in
    (* put some useful stuff in the context to help eauto *)
    assert (typing nil th_prog (exp_fun (TermInterp rho a 2))
                                        (typ_arr A B))
        as HTyp2 by eauto; 
    assert (forall j, j < k -> interp (j,typ_arr A B,Comp) th_prog 
                                      (TermInterp rho (exp_fun a) 0))
        as IHInterp by (intros; eauto);
    unfold_interp; intros; autorewrite with steps in *; rewrite open_open_swap;
    eauto 10
  | [ |- interp (_, typ_arr _ _, _) th_log 
                (exp_lam (TermInterp ?rho ?b 1)) ] =>
    (* TLam *)  (* CHECKED *)
    unfold_interp;
    exists (exp_lam (TermInterp rho b 1)); split; eauto;
    unfold_interp;
    match goal with
    | [ |- forall _ _, _ <= _ -> _ -> _ ] =>
      let Hle := fresh "Hle" in intros ? ? Hle; intros;
      destruct (le_lt_or_eq _ _ Hle); subst; eauto
    end
  | [ HF : FO ?A, HV : val ?v |- interp _ _ (TermInterp ?rho ?v 0) ] =>
    (* FOVal case *)
    unfold_interp; unfold_interp_hyps;
    exists (TermInterp rho v 0); split; eauto
  | [ |- interp (_,typ_nat,_) ?t exp_zero ] =>
    (* TZero case *)
    unfold_interp; match goal with
    | [ |- exists _, _ /\ _ ] => exists exp_zero; split
    | [ |- forall _ _, _ -> _ ] => intros
    end; try unfold_interp; eauto
  | [ IHtyp : interp (?k,_,_) _ ?a 
     |- interp (?k,_,_) th_prog (exp_unroll ?a) ] =>
    (* Unroll *)
     unfold_interp; unfold_interp_hyps; intros ? ? ? HMS ?;
     mstep_val_inversion; apply IHtyp in HMS; unfold_interp_hyps; eauto
  | [ IHtyp : interp (?k,?A,_) _ (TermInterp ?rho ?a 0)
     |- interp (?k,typ_at ?A _,_) _ (TermInterp ?rho ?a 0) ] =>
    (* TBoxL, TBoxLV, TBoxP *)
    unfold_interp; unfold_interp_hyps;
    try match goal with
    | [ H : val ?a |- exists _, _ /\ _ ] =>
      exists (TermInterp rho a 0); unfold_interp; split; eauto
    | [ H : astep ?a ?v |- exists _, _ /\ _ ] =>
      exists v; unfold_interp; split; eauto
    | [ |- forall _ _, _ -> mstep _ _ _ -> val _ -> _ ] =>
      intros ? ? ? HMS HVal; mstep_val_inversion; instantiate;
      subst; try (apply IHtyp in HMS); instantiate; auto;
      unfold_interp; eauto
    end
  | [ IHtyp : interp (?k,_,_) _ (TermInterp ?rho ?a 0)
     |- interp (?k,_,_) _ (?P (TermInterp ?rho ?a 0)) ] =>
    (* TSucc, TRoll, TInl, TInr *)
    unfold_interp; unfold_interp in IHtyp;
    match goal with
    | [ H : val ?a |- exists _, _ /\ _ ] =>
      exists (P (TermInterp rho a 0)); unfold_interp; split; eauto
    | [ H : astep ?a ?v |- exists _, _ /\ _ ] =>
      exists (P v); unfold_interp; split; eauto
    | [ |- forall _ _, _ -> mstep _ (P _) _ -> val _ -> _ ] =>
      intros ? ? ? HMS HVal; mstep_val_inversion; instantiate;
      subst; try (apply IHtyp in HMS); instantiate; auto;
      unfold_interp; eauto
    end
  | [ IHa : interp _ _ (TermInterp _ ?a _),
      IHb1 : interp _ _ (TermInterp _ ?b1 _),
      IHb2 : forall _, wfTermEnv ?k (_ :: ?G) _ 
                    -> interp _ _ (TermInterp _ ?b2 _)
     |- interp _ _ (exp_natcase (TermInterp _ ?a _) (TermInterp _ ?b1 _)
                              (TermInterp _ ?b2 _)) ] =>
    (* TCase *)
    unfold_interp; unfold_interp_hyps;
    match goal with
    | [ H : looks_like_nat ?v,
        ASb1 : astep (TermInterp _ b1 _) ?v1
       |- exists v', _ /\ _ ] =>
      destruct v; simpl in H; try inversion H; subst;
      match goal with
      | [ HAS : astep (TermInterp _ a _) exp_zero |- _ ] => 
        exists v1; split; ASCongSolver; eauto
      | [ HAS : astep (TermInterp _ a _) (exp_succ ?a') |- _ ] =>
        let Hwf := fresh "Hwf" in
        assert (wfTermEnv k ((th_log,typ_nat) :: G) (v0 :: rho))
         as Hwf by eauto;
        apply IHb2 in Hwf; unfold_interp_hyps; autorewrite with steps in *;
        match goal with
        | [ HI : astep (TermInterp _ ({_ ~> _}_) _) ?v2 |- _ ] =>
          exists v2; split; ASCongSolver; eauto
        end
      end 
    | [ |- forall _ _, _ <= _ -> mstep _ _ _ -> _ -> interp _ _ _ ] =>
      intros ? ? ? HMS ?;
      apply mstep_natcase_val_inversion in HMS; auto;
      destruct HMS as [j1 [j2 [? [ [? ?] | [? [? [? ?]]] ]]]]; subst;
      try match goal with
      | [ HMSa : mstep _ (TermInterp _ a _) exp_zero,
          HMSb1 : mstep _ (TermInterp _ b1 _) _ |- _ ] =>
        apply IHb1 in HMSb1; auto
      | [ HMSa : mstep _ (TermInterp _ a _) (exp_succ ?va),
          HMSb2 : mstep _ ({0 ~> ?va} (TermInterp _ b2 _)) _ |- _ ] =>
        apply IHa in HMSa; auto; unfold_interp_hyps;
        let IHApplied := fresh "IHApplied" in
        assert (interp (k - S j1, A, Comp) th_prog 
                       (TermInterp (x :: rho) b2 0)) as IHApplied by eauto;
        unfold_interp_hyps;
        replace (k - (1 + j1 + j2)) with (k - S j1 - j2) by omega;
        apply IHApplied; eauto
      end
    end
  (* TSumCase *)
  | [ IHbl : forall _, wfTermEnv ?k ((?t',?A) :: ?G) _
                   -> interp _ _ (TermInterp _ ?bl _),
      IHbr : forall _, wfTermEnv ?k ((?t',?B) :: ?G) _
                   -> interp _ _ (TermInterp _ ?br _)
      |- interp _ _ (exp_sumcase (TermInterp ?rho ?a _)
                        (TermInterp _ ?bl _) (TermInterp _ ?br _)) ] =>
    unfold_interp; unfold_interp_hyps; autorewrite with steps in *;
    ASCongSolver;
    try match goal with
    | [ |- exists v', _ /\ interp (k,?C,Val) ?t v' ] =>
      let Hwf := fresh "Hwf" in
      match goal with
      | [ Hastepl : astep (TermInterp _ a 0) (exp_inl ?vl) |- _ ] =>
        assert (wfTermEnv k ((t',A) :: G) (vl :: rho)) as Hwf by eauto;
        apply IHbl in Hwf; clear IHbl IHbr
      | [ Hastepr : astep (TermInterp _ a 0) (exp_inr ?vr) |- _ ] =>
        clear IHbl;
        assert (wfTermEnv k ((t',B) :: G) (vr :: rho)) as Hwf by eauto;
        apply IHbr in Hwf; clear IHbr IHbl
      end; unfold_interp_hyps; autorewrite with steps in *;
      match goal with
      | [ Hwf : interp (k, C, Val) _ ?v |- _ ] => 
        exists v; split; ASCongSolver; eauto
      end
    | [ |- forall _ _, _ <= _ -> mstep _ _ _ -> _ -> interp _ _ _ ] =>
      intros ? ? ? HMS ?;
      apply mstep_sumcase_val_inversion in HMS; auto;
      destruct HMS as [j1 [j2 [? [? [? [ [? ?] | [? ?] ]]]]]]; subst;
      match goal with
      | [ IHA : forall _ _, _ ->  mstep _ (TermInterp _ a _) _ -> _,
          HStepA : mstep ?j1 (TermInterp rho a 0) ?v |- _ ] =>
        apply IHA in HStepA; auto; clear IHA; unfold_interp_hyps
      end;
      match goal with
      | [ Hstepcase : mstep j2 ({_ ~> _}(TermInterp rho ?blr _)) _ |- _ ] => 
        assert (interp (k - S j1, C, Comp) th_prog (TermInterp (x :: rho) blr 0))
            as IHApplied by (eapply IHWf; eauto);
        unfold_interp_hyps;
        replace (k - (1 + j1 + j2)) with (k - S j1 - j2) by omega;
        eapply IHApplied; eauto
      end
    end
  | [ IHTyp : interp (_, typ_at ?A ?t', _) _ _ 
     |- interp (_, ?A, _) ?t' _ ] =>
    (* TUnboxVal *)
    unfold_interp; unfold_interp_hyps; intros;
    (* Idea: we now have a value in the context and a bunch of
       assumptions that that value multisteps to things.  Eliminate
       those assumptions to get proofs that the thing stepped to was
       actually the val.  Substitute, unfold, and then our assumptions
       look like our conclusion.
     *)
    repeat (match goal with
    (* Begin by turning ASteps into MSteps for uniformity *)
    | [ HV : val ?a, HAS : astep (TermInterp ?rho ?a 0) _ |- _ ] =>
      inversion HAS; subst; clear HAS
    end);
    repeat (match goal with
    (* clear out MSteps because they are from values *)
    | [ HV : val ?a, HMS : mstep _ (TermInterp ?rho ?a 0) ?v |- _ ] =>
      assert (TermInterp rho a 0 = v)
        by (inversion HMS; subst; eauto;
            match goal with
            | [ HS : step (TermInterp rho a 0) _ |- _ ] =>
              apply values_dont_step in HS; [ inversion HS | eauto ]
            end);
      subst; clear HMS
    end);
    try match goal with
    (* instantiate hyps that depend on mstepping from a value *)
    | [ IHT : forall _ _, _ -> mstep _ (TermInterp ?rho ?a _) _ -> _ -> _
       |- _ ] => 
      let HAS := fresh "HAS" in
      assert (mstep 0 (TermInterp rho a 0) (TermInterp rho a 0))
             as HAS by eauto;
      apply IHT in HAS; [| eauto | eauto]
    end;
    (* declare victory *)
    unfold_interp_hyps; eauto; autorewrite with  steps in *; eauto
  end.
Qed.


Theorem normalization : forall a A,
     typing nil th_log a A
  -> exists v, val v /\ astep a v.
Proof.
  intros a A HTyp.
  apply interp_soundness with (k := 0) (rho := nil) in HTyp;
  try unfold_interp_hyps; eauto.
Qed.

Theorem false_proof_not_val : forall G t a A,
     typing G t a A
  -> G = nil
  -> (   A = (typ_mu (typ_var 0))
      \/ exists t, Boxed (typ_mu (typ_var 0)) t A)
  -> val a -> False.
Proof.
  induction 1; intros; subst; destruct_things; subst; eauto;
  val_inversion; try match goal with
  | [ HEq : _ = typ_mu _ |- _ ] => inversion HEq; subst; eauto
  end;
  try solve [match goal with
  | [ H : ?G1 ++ _ :: _ = nil |- _ ] =>
    destruct G1; inversion H
  | [ H : FO (typ_mu _) |- _ ] =>
    inversion H
  | [ H : Boxed (typ_mu _) _ (typ_at _ _) |- _ ] =>
    inversion H; subst; eauto
  end].
Qed.

(* without a proper False type and elimination principle, this
   consistency theorem is a little boring. *)
Theorem consistency : forall a,
     typing nil th_log a (typ_mu (typ_var 0))
  -> False.
Proof.
  intros a HTyp;
  destruct (normalization _ _ HTyp) as [x [? ?]];
  edestruct false_proof_not_val with (a := x); eauto.
Qed.
