Require Import List.
Require Import Arith.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_reduction_basics.
Require Import step_substitution.


(* Some ltac that should probably be moved *)
Ltac destruct_ih :=
  match goal with
  [ IH : forall _ _ _, _ = _ -> _ = _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ _ eq_refl eq_refl); clear IH
  | [ IH : forall _ _, ?P _ _ = ?P _ _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ eq_refl); clear IH
  | [ IH : forall _ _ _ _, ?P _ = ?P _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ _ _ eq_refl); clear IH
  | [ IH : forall _ _, ?P _ = ?P _ -> ?Q _ = ?Q _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ eq_refl eq_refl); clear IH
  | [ IH : forall _, ?P _ = ?P _ -> ?A = ?A -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ eq_refl eq_refl); clear IH
  | [ IH : forall _, ?P _ = ?P _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ eq_refl); clear IH
  | [ IH : ?P = ?P -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH eq_refl); clear IH
  | [ IH : forall _ _ _, ?P _ _ _ = ?P _ _ _ -> _ |- _ ] =>
    let IH' := fresh IH in
    assert (IH' := IH _ _ _ eq_refl); clear IH
  end; destruct_things.


Theorem TSub_gen_l : forall G t a A,
  typing G th_log a A -> typing G t a A.
Proof.  intros G t a A H; destruct t; eauto.  Qed.

Theorem TSub_gen_p : forall G t a A,
  typing G t a A -> typing G th_prog a A.
Proof. intros ? t; destruct t; eauto.  Qed.
Hint Immediate TSub_gen_l TSub_gen_p.

Ltac fo_inversion :=
  try match goal with
  | [ H : FO (typ_arr _ _ _) |- _ ] => inversion H
  | [ H : FO (typ_mu _) |- _ ] => inversion H
  | [ H : FO (typ_sum _ _) |- _ ] => inversion H; subst
  end.

Hint Extern 0 => match goal with
  | [ H : FO (typ_arr _ _) |- _ ] => inversion H
  | [ H : FO (typ_var _) |- _ ] => inversion H
  | [ H : FO (typ_mu _) |- _ ] => inversion H
  end.

Theorem fo_inv_sum_l : forall A B, FO (typ_sum A B) -> FO A.
Proof. intros; fo_inversion; auto. Qed.

Theorem fo_inv_sum_r : forall A B, FO (typ_sum A B) -> FO B.
Proof. intros; fo_inversion; auto. Qed.
Hint Immediate fo_inv_sum_l fo_inv_sum_r.

Hint Extern 0 => match goal with
  | [ H : th_log = th_prog |- _ ] => inversion H
  end.

(******************************************************************)
(* Some lemmas about changing consistency classifiers in the ctxt *)

(**** weakening ***)
Theorem theta_weakening' : forall G t a A,
     typing G t a A
  -> forall G1 B G2, G = G1 ++ (th_prog,B) :: G2
      -> typing (G1 ++ (th_log,B) :: G2) t a A.
Proof.
  induction 1; intros; subst; eauto; inst_equalities; eauto.  
  (* Only the var case remains *)
  match goal with
  | [ HEq : ?G1 ++ _ :: _ = ?G1' ++ _ :: _ |- _ ] =>
    destruct_compares (length G1) (length G1');
    invert_list_equalities; subst; eauto
  end; find_cons;
  [ eauto using TSub
  | rewrite <- app_assoc, <- app_comm_cons; simpl; eauto
  | rewrite app_comm_cons, app_assoc;
    apply TVar; autorewrite with steps; simpl; eauto ].
  Qed.

Theorem theta_weakening : forall G1 B G2 t a A,
     typing (G1 ++ (th_prog,B) :: G2) t a A
  -> typing (G1 ++ (th_log,B) :: G2) t a A.
Proof. intros; eauto using theta_weakening'. Qed.

Theorem theta_weakening_cons : forall B G t a A,
     typing ((th_prog,B) :: G) t a A
  -> typing ((th_log,B) :: G) t a A.
Proof.
  intros B G t a A H; apply theta_weakening with (G1 := nil) in H;
  simpl in *; eauto.
Qed.

(**** strengthening ***)
Theorem theta_strengthening_fo' : forall G t a A, typing G t a A
  -> forall G1 B G2, G = G1 ++ (th_log,B) :: G2 -> FO B
      -> typing (G1 ++ (th_prog,B) :: G2) t a A.
Proof.
  induction 1; intros; subst; eauto; inst_equalities; eauto.

  (* Only the var case remains *)
  match goal with
  | [ HEq : ?G1 ++ _ :: _ = ?G1' ++ _ :: _ |- _ ] =>
    destruct_compares (length G1) (length G1');
    invert_list_equalities; subst; eauto
  end; find_cons; 
  [ eauto
  | rewrite <- app_assoc, <- app_comm_cons; simpl; eauto
  | rewrite app_comm_cons, app_assoc;
    apply TVar; autorewrite with steps; simpl; eauto ].
Qed.

Theorem theta_strengthening_fo : forall G1 B G2 t a A,
     typing (G1 ++ (th_log,B) :: G2) t a A -> FO B
  -> typing (G1 ++ (th_prog,B) :: G2) t a A.
Proof.  intros; eauto using theta_strengthening_fo'.  Qed.

Theorem theta_strengthening_fo_cons : forall B G t a A,
     typing ((th_log,B) :: G) t a A
  -> FO B
  -> typing ((th_prog,B) :: G) t a A.
Proof.
  intros B G t a A HT HF;
  apply theta_strengthening_fo with (G1 := nil) in HT;
  simpl in *; eauto.
Qed.
Hint Immediate theta_strengthening_fo theta_strengthening_fo_cons.


(*************************)
(* Inversion for lambdas *)
Theorem typ_lam_inversion' : forall G t a A,
     typing G t a A
  -> forall b, a = exp_lam b
     ->    (exists B1:typ, exists B2:typ,
               FO B1
            /\ typing ((th_log,B1) :: G) th_log b B2
            /\ (   (A = typ_arr B1 B2)
                \/ (exists t', Boxed (typ_arr B1 B2) t' A))).
Proof.
  induction 1; intros;
  try match goal with
  | [ H : _ = exp_lam _  |- _] => inversion H; subst; clear H
  end; try destruct_ih; subst; try match goal with
  | [ H : typ_at _ _ = typ_arr _ _ |- _ ] => inversion H
  | [ H : Boxed (typ_arr _ _) _ (typ_at _ _) |- _ ] => inversion H; subst; eauto 8
  end; eauto 8.
Qed.

Theorem typ_lam_inversion : forall G t b B,
     typing G t (exp_lam b) B
  -> exists B1:typ, exists B2:typ,
         FO B1
      /\ typing ((th_log,B1) :: G) th_log b B2
      /\ (   (B = typ_arr B1 B2)
          \/ (exists t', Boxed (typ_arr B1 B2) t' B)).
Proof.
  intros; eauto using typ_lam_inversion'.
Qed.

Theorem typ_lam_inversion_arr : forall G t b A B,
     typing G t (exp_lam b) (typ_arr A B)
  -> FO A /\ typing ((th_log,A) :: G) th_log b B.
Proof.
  intros; match goal with
  | [ H : typing _ _ _ _ |- _ ] =>
    apply typ_lam_inversion in H; destruct_things; 
    try match goal with
    | [ HEq : _ = typ_arr _ _ |- _ ] => inversion HEq; subst
    end; eauto
  end.
Qed.
Hint Immediate typ_lam_inversion_arr.
(*************************)
(*************************)


(***************************)
(* Inversion for recursion *)
Theorem typ_fun_inversion' : forall G t a A,
     typing G t a A
  -> forall b, a = exp_fun b
     -> (exists B1:typ, exists B2:typ,
            FO B1
         /\ typing ((th_prog, typ_arr B1 B2) :: (th_prog,B1) :: G) 
                    th_prog b B2
         /\ (   (A = typ_arr B1 B2 /\ t = th_prog)
             \/ Boxed (typ_arr B1 B2) th_prog A)).
Proof.
  induction 1; intros;
  try match goal with
  | [ H : _ = exp_fun _  |- _] => inversion H; subst; clear H
  end; try destruct_ih; subst;
  try match goal with
  | [ H : typ_at _ _ = typ_arr _ _ |- _ ] => inversion H
  | [ H : Boxed (typ_arr _ _) _ (typ_at _ _) |- _ ] => inversion H; subst; eauto 8
  end; eauto 8.
Qed.

Theorem typ_fun_inversion : forall G t b B,
     typing G t (exp_fun b) B
  -> (exists B1:typ, exists B2:typ,
         FO B1
      /\ typing ((th_prog, typ_arr B1 B2) :: (th_prog,B1) :: G) 
                 th_prog b B2
      /\ (   (B = typ_arr B1 B2 /\ t = th_prog)
          \/ Boxed (typ_arr B1 B2) th_prog B)).
Proof. intros; eauto using typ_fun_inversion'. Qed.

Theorem typ_fun_inversion_arr : forall G t b A B,
     typing G t (exp_fun b) (typ_arr A B)
  ->    FO A
     /\ t = th_prog
     /\ typing ((th_prog, typ_arr A B) :: (th_prog,A) :: G) 
               th_prog b B.
Proof.
  intros; match goal with
  | [ H : typing _ _ _ _ |- _ ] =>
    apply typ_fun_inversion in H; destruct_things;
    try match goal with
    | [ HEq : _ = typ_arr _ _ |- _ ] => inversion HEq; subst
    end; eauto
  end.
Qed.
Hint Immediate typ_fun_inversion typ_fun_inversion_arr.
(*************************)
(*************************)

(*
(*********************)
(* Inversion for app *)
Theorem typ_app_inversion' : forall G t a B,
     typing G t a B
  -> forall b1 b2, a = exp_app b1 b2
     -> exists A:typ, exists t':th,
            typing G t b1 (typ_arr A B)
         /\ typing G t b2 A.
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_app _ _ |- _ ] => inversion H; subst; clear H
  end; try destruct_ih; subst; eauto 8.
 match goal with
  | [ IH : forall _ _, exp_app _ _ = exp_app _ _ -> _ |- _ ] =>
    destruct (IH _ _ eq_refl) as [? [? [? ?]]]; clear IH;
    subst; eauto 6
  end; val_inversion.
Qed.

Theorem typ_app_inversion : forall G t a b B,
     typing G t (exp_app a b) B
  -> exists A:typ, exists t':th,
         typing G t a (typ_arr A t' B)
      /\ typing G t (exp_box b) (typ_at A t').
Proof.  intros; eauto using typ_app_inversion'.  Qed.
*)

(**********************)
(* Inversion for roll *)
Theorem typ_roll_inversion' : forall G t a A, typing G t a A
  -> forall b, a = exp_roll b
     -> exists B,
            typing G th_prog b ({0 ~~> typ_mu B} B)
         /\ (   A = typ_mu B
             \/ exists t', Boxed (typ_mu B) t' A).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_roll _ |- _ ] => inversion H; subst; clear H
  end; try destruct_ih; subst; eauto 8;
  match goal with
  | [ H : typ_at _ _ = typ_mu _ |- _ ] => inversion H
  | [ H : Boxed (typ_mu _) _ (typ_at _ _) |- _ ] => inversion H; subst; eauto 8
  | [ H : typing _ ?t _ ({_ ~~> typ_mu _} _) |- _ ] =>
    is_var t; destruct t; eauto 8
  end.
Qed.

Theorem typ_roll_inversion : forall G t b A,
     typing G t (exp_roll b) A
  -> exists B,
         typing G th_prog b ({0 ~~> typ_mu B} B)
      /\ (   A = typ_mu B
          \/ exists t', Boxed (typ_mu B) t' A).
Proof.  intros; eauto using typ_roll_inversion'.  Qed.

Theorem typ_roll_inversion_mu: forall G t b B,
     typing G t (exp_roll b) (typ_mu B)
  -> typing G th_prog b ({0 ~~> typ_mu B} B).
Proof.
  intros G t b B HTyp; apply typ_roll_inversion in HTyp;
  destruct_things;
  try match goal with
  | [ H : typ_mu _ = typ_mu _ |- _ ] => inversion H; subst
  end; eauto.
Qed.
Hint Immediate typ_roll_inversion typ_roll_inversion_mu.
(*********************)
(*********************)

(*
(************************)
(* Inversion for unroll *)
Theorem typ_unroll_inversion' : forall G t a A, typing G t a A
  -> forall b, a = exp_unroll b
     -> exists B,
            t = th_prog
         /\ A = {0 ~~> typ_mu B} B
         /\ typing G th_prog b (typ_mu B).
Proof.
  induction 1; intros; try match goal with
  | [ H : _ = exp_unroll _ |- _ ] => inversion H; subst; clear H
  end; eauto 6;
  match goal with
  | [ IH : forall _, exp_unroll _ = exp_unroll _ -> _ |- _ ] =>
    destruct (IH _ eq_refl) as [? [? [? ?]]]; clear IH;
    subst; eauto 6
  end; val_inversion. 
Qed.

Theorem typ_unroll_inversion : forall G t b A,
     typing G t (exp_unroll b) A
  -> exists B,
         t = th_prog
      /\ A = {0 ~~> typ_mu B} B
      /\ typing G th_prog b (typ_mu B).
Proof.  intros; eauto using typ_unroll_inversion'.  Qed.
Hint Immediate typ_unroll_inversion.
(*********************)
(*********************)


(**********************)
(* Inversion for natcase *)
Theorem typ_natcase_inversion' : forall G t a A,
     typing G t a A
  -> forall b b1 b2, a = exp_natcase b b1 b2
      ->    typing G t b typ_nat
         /\ typing G t b1 A
         /\ typing ((t,typ_nat) :: G) t b2 A.
Proof.
  induction 1; intros; match goal with
  | [ HEq : _ = exp_natcase _ _ _ |- _ ] => inversion HEq; subst
  end; val_inversion; eauto; match goal with
  | [ IH : forall _ _ _, exp_natcase _ _ _ = exp_natcase _ _ _ -> _ |- _ ] =>
  destruct (IH _ _ _ eq_refl) as [? [? ?]]; eauto 10
  end.
Qed.

Theorem typ_natcase_inversion : forall G t b b1 b2 A,
     typing G t (exp_natcase b b1 b2) A
  ->    typing G t b typ_nat
     /\ typing G t b1 A
     /\ typing ((t,typ_nat) :: G) t b2 A.
Proof.  intros; eauto using typ_natcase_inversion'.  Qed.
Hint Immediate typ_natcase_inversion.
(*********************)
(*********************)

(*************************)
(* Inversion for sumcase *)
Theorem typ_sumcase_inversion' : forall G t a A,
     typing G t a A
  -> forall b b1 b2, a = exp_sumcase b b1 b2
      -> exists B1, exists B2, exists t',
            typing G t (exp_box b) (typ_at (typ_sum B1 B2) t')
         /\ typing ((t',B1)::G) t b1 A
         /\ typing ((t',B2)::G) t b2 A.
Proof.
  induction 1; intros; match goal with
  | [ HEq : _ = exp_sumcase _ _ _ |- _ ] => inversion HEq; subst
  end; val_inversion; eauto 6;
  match goal with
  | [ IH : forall _ _ _, exp_sumcase _ _ _ = exp_sumcase _ _ _ -> _ |- _ ] =>
  destruct (IH _ _ _ eq_refl) as [? [? [? [? [? ?]]]]]; eauto 10
  end.
Qed.

Theorem typ_sumcase_inversion : forall G t b b1 b2 A,
     typing G t (exp_sumcase b b1 b2) A
  -> exists B1, exists B2, exists t',
            typing G t (exp_box b) (typ_at (typ_sum B1 B2) t')
         /\ typing ((t',B1)::G) t b1 A
         /\ typing ((t',B2)::G) t b2 A.
Proof.  intros; eauto using typ_sumcase_inversion'.  Qed.
Hint Immediate typ_sumcase_inversion.
(*********************)
(*********************)
*)

(**********************)
(* Inversion for succ *)
Theorem typ_succ_inversion' : forall G t a A,
     typing G t a A
  -> forall a', a = exp_succ a' 
       ->   (   A = typ_nat
             /\ typing G t a' typ_nat)
         \/ (exists t',
                Boxed typ_nat t' A
             /\ typing G t' a' typ_nat).
Proof.

  induction 1; intros; match goal with
  | [ HEq : _ = exp_succ _ |- _ ] => inversion HEq; subst
  end; val_inversion; try destruct_ih; subst; eauto 8;
  match goal with
  | [ H : typ_at _ _ = typ_nat |- _ ] => inversion H
  | [ H : Boxed typ_nat _ (typ_at _ _) |- _ ] => inversion H; subst; eauto 8
  end.
Qed.

Theorem typ_succ_inversion : forall G t a A,
     typing G t (exp_succ a) A
 ->   (   A = typ_nat
       /\ typing G t a typ_nat)
   \/ (exists t',
          Boxed typ_nat t' A
       /\ typing G t' a typ_nat).
Proof.  intros; eauto using typ_succ_inversion'.  Qed.
Hint Immediate typ_succ_inversion.

Theorem typ_succ_inversion_nat : forall G t a,
     typing G t (exp_succ a) typ_nat
  -> typing G t a typ_nat.
Proof.
  intros G t a H; apply typ_succ_inversion in H;
  destruct_things; eauto.
Qed.

(*********************)
(*********************)

(*********************)
(* Inversion for inl *)
Theorem typ_inl_inversion' : forall G t a A,
     typing G t a A
  -> forall a', a = exp_inl a'
       -> (exists B1, exists B2, 
               (A = typ_sum B1 B2 /\ typing G t a' B1)
            \/ exists t',
                 Boxed (typ_sum B1 B2) t' A /\ typing G t' a' B1).
Proof.
  induction 1; intros; match goal with
  | [ HEq : _ = exp_inl _ |- _ ] => inversion HEq; subst
  end; val_inversion; try destruct_ih; subst; eauto 8 using TFOVal;
  match goal with
  | [ H : typ_at _ _ = typ_sum _ _ |- _ ] => inversion H
  | [ H : Boxed (typ_sum _ _) _ (typ_at _ _) |- _ ] => inversion H; subst; eauto 8
  end.
Qed.

Theorem typ_inl_inversion : forall G t a A,
     typing G t (exp_inl a) A
  -> (exists B1, exists B2, 
          (A = typ_sum B1 B2 /\ typing G t a B1)
       \/ exists t',
            Boxed (typ_sum B1 B2) t' A /\ typing G t' a B1).
Proof.  intros; eauto using typ_inl_inversion'.  Qed.

Theorem typ_inl_inversion_sum : forall G t a A1 A2,
     typing G t (exp_inl a) (typ_sum A1 A2)
  -> typing G t a A1.
Proof.
  intros G t a A1 A2 H; apply typ_inl_inversion in H;
  destruct_things; eauto;
  match goal with
  | [ HEq : typ_sum _ _ =  typ_sum _ _ |- _ ] =>
    inversion HEq; subst; eauto
  end.
Qed.


(*********************)
(*********************)

(*********************)
(* Inversion for inr *)
Theorem typ_inr_inversion' : forall G t a A,
     typing G t a A
  -> forall a', a = exp_inr a'
       -> (exists B1, exists B2, 
               (A = typ_sum B1 B2 /\ typing G t a' B2)
            \/ exists t',
                 Boxed (typ_sum B1 B2) t' A /\ typing G t' a' B2).
Proof.
  induction 1; intros; match goal with
  | [ HEq : _ = exp_inr _ |- _ ] => inversion HEq; subst
  end; val_inversion; try destruct_ih; subst; eauto 8 using TFOVal;
  match goal with
  | [ H : typ_at _ _ = typ_sum _ _ |- _ ] => inversion H
  | [ H : Boxed (typ_sum _ _) _ (typ_at _ _) |- _ ] => inversion H; subst; eauto 8
  end.
Qed.

Theorem typ_inr_inversion : forall G t a A,
     typing G t (exp_inr a) A
  -> (exists B1, exists B2, 
          (A = typ_sum B1 B2 /\ typing G t a B2)
       \/ exists t',
            Boxed (typ_sum B1 B2) t' A /\ typing G t' a B2).
Proof.  intros; eauto using typ_inr_inversion'.  Qed.

Theorem typ_inr_inversion_sum : forall G t a A1 A2,
     typing G t (exp_inr a) (typ_sum A1 A2)
  -> typing G t a A2.
Proof.
  intros G t a A1 A2 H; apply typ_inr_inversion in H;
  destruct_things; eauto;
  match goal with
  | [ HEq : typ_sum _ _ =  typ_sum _ _ |- _ ] =>
    inversion HEq; subst; eauto
  end.
Qed.
(*********************)
(*********************)
Ltac untyped a :=
  match goal with
  | [ H : typing _ _ a _ |- _ ] => fail 1
  | _ => idtac
  end.

Ltac typing_inversion :=
  match goal with
  | [ H : typing _ _ (exp_lam ?b) (typ_arr _ _) |- _] =>
    untyped b; let H' := fresh H in
    assert (H' := typ_lam_inversion_arr _ _ _ _ _ H);
    destruct_things
  | [ H : typing _ _ (exp_fun ?b) (typ_arr _ _) |- _] =>
    untyped b; let H' := fresh H in
    assert (H' := typ_fun_inversion_arr _ _ _ _ _ H);
    destruct_things
  | [ H : typing _ _ (exp_succ _) typ_nat |- _] =>
    apply typ_succ_inversion_nat in H
  | [ H : typing _ _ (exp_inl ?b) (typ_sum _ _) |- _] =>
    untyped b; let H' := fresh H in
    assert (H' := typ_inl_inversion_sum _ _ _ _ _ H);
    destruct_things
  | [ H : typing _ _ (exp_inl ?b) ?B |- _ ] =>
    untyped b; let H' := fresh H in
    assert (H' := typ_inl_inversion _ _ _ _ H);
    destruct_things; subst;
    try match goal with
    | [ HEq : typ_at _ _ = typ_sum _ _ |- _ ] => inversion HEq
    | [ HB : Boxed (typ_sum _ _) _ (typ_at (typ_sum _ _) _) |- _ ] =>
      inversion HB; subst
    end
  | [ H : typing _ _ (exp_inr ?b) (typ_sum _ _) |- _] =>
    untyped b; let H' := fresh H in
    assert (H' := typ_inr_inversion_sum _ _ _ _ _ H);
    destruct_things
  | [ H : typing _ _ (exp_inr ?b) ?B |- _ ] =>
    untyped b; let H' := fresh H in
    assert (H' := typ_inr_inversion _ _ _ _ H);
    destruct_things; subst;
    try match goal with
    | [ HEq : typ_at _ _ = typ_sum _ _ |- _ ] => inversion HEq
    | [ HB : Boxed (typ_sum _ _) _ (typ_at (typ_sum _ _) _) |- _ ] =>
      inversion HB; subst
    end
  | [ H : typing _ _ (exp_roll ?b) (typ_mu _) |- _] =>
    untyped b; let H' := fresh H in
    assert (H' := typ_roll_inversion_mu _ _ _ _ H);
    destruct_things
  end.

(* I'm going to do the lam and fun cases ahead of time for clarity *)
Theorem typ_fun_inversion_subst : forall G t b A1 A2,
     typing G t (exp_fun b) (typ_arr A1 A2)
  -> forall v, val v -> typing G t v A1 
           -> typing G t ({0 ~> v} ({0 ~> exp_fun (lift_e b 1 2)} b))
                     A2.
Proof.
  intros; typing_inversion; subst; extract_lift_fun; eauto 10.
Qed.  

Theorem typ_lam_inversion_subst : forall G t b A1 A2,
     typing G t (exp_lam b) (typ_arr A1 A2)
  -> forall v, val v -> typing G t v A1 
           -> typing G t ({0 ~> v} b) A2.
Proof.
  intros; typing_inversion; eauto 8.
Qed.

Theorem preservation : forall G t a A,
     typing G t a A
  -> forall a', step a a' -> typing G t a' A.
Proof.
  induction 1; intros; eauto; try step_val_elim;
  match goal with
  | [ HS : step _ _ |- _ ] => inversion HS; subst
  end; eauto; try match goal with
  | [ |- typing _ _ ({_ ~> _} ({_ ~> exp_fun _} _)) _ ] =>
    eauto using typ_fun_inversion_subst
  | [ H : typing _ _ (exp_lam ?b) _ |- typing _ _ ({_ ~> _}?b) _ ] =>
    eauto using typ_lam_inversion_subst
  end; try typing_inversion; eauto using TSub.
Qed.

Theorem mpreservation : forall G t a b A n,
  typing G t a A -> mstep n a b -> typing G t b A.
Proof.
  induction 2; eauto using preservation.
Qed.

Theorem apreservation : forall G t a b A,
  typing G t a A -> astep a b -> typing G t b A.
Proof.
  inversion 2; eauto using mpreservation.
Qed.

Hint Immediate preservation  mpreservation apreservation.
