Require Import List.
Require Import Arith.

Require Import step_defs.
Require Import step_db_basics.
Require Import step_reduction_basics.

Theorem closed_no_var' : forall G t a A,
     typing G t a A
  -> forall x, G = nil -> a = exp_var x -> False.
Proof.
  induction 1; intros; subst; eauto;
  try match goal with
  | [ H : ?G1 ++ _ :: _ = nil |- _ ] => destruct G1; simpl in H; inversion H
  | [ H : _ = _ |- _ ] => solve [inversion H]
  end.
Qed.

Theorem closed_no_var : forall t n A,
  typing nil t (exp_var n) A -> False.
Proof.
  eauto using closed_no_var'.
Qed.
Hint Resolve closed_no_var.


Hint Extern 0 =>
  match goal with
  | [ HEq : ?G ++ _ :: _ = nil |- _ ] => destruct G; inversion HEq
  end.

Theorem canonical_forms_nat' : forall G t a A,
     typing G t a A 
  -> val a -> G = nil
  -> (A = typ_nat \/ (exists t', Boxed typ_nat t' A))
  -> a = exp_zero \/ (exists a', a = exp_succ a').
Proof.
  induction 1; intros; subst; eauto;
  val_inversion; destruct_things; subst; eauto;
  try solve [match goal with
  | [ H : _ = _ |- _ ] => inversion H
  end];
  match goal with
  | [ IH : _ -> _ -> _ -> ?P,
      H : Boxed _ _ _ |- ?P ] =>
    apply IH; eauto; inversion H; subst; eauto
  end.
Qed.

Theorem canonical_forms_nat_closed : forall t a,
     typing nil t a typ_nat -> val a
  -> a = exp_zero \/ exists a', a = exp_succ a'.
Proof.
  intros t a HTyp HVal;
  destruct (canonical_forms_nat' _ _ _ _ HTyp); eauto.
Qed.

Theorem canonical_forms_arrow' : forall G t a A,
     typing G t a A 
  -> val a -> forall B1 B2,
     (A = typ_arr B1 B2 \/ exists t', Boxed (typ_arr B1 B2) t' A)
  ->    (exists b, a = exp_lam b)
     \/ (exists b, a = exp_fun b)
     \/ (exists n, a = exp_var n).
Proof.
  induction 1; intros; subst; eauto;
  val_inversion; destruct_things; subst; eauto;
  try solve [match goal with
  | [ H : _ = _ |- _ ] => inversion H
  end]; try match goal with
  | [ IH : _ -> forall _ _, _ -> ?P,
      H : Boxed _ _ _ |- ?P ] => 
    eapply IH; eauto; inversion H; subst; eauto
  end.
Qed.

Theorem canonical_forms_arrow_closed : forall t a A B,
     typing nil t a (typ_arr A B) -> val a
  ->    (exists b, a = exp_lam b)
     \/ (exists b, a = exp_fun b).
Proof.
  intros t a A B HTyp HVal;
  edestruct (canonical_forms_arrow' _ _ _ _ HTyp); eauto;
  destruct_things;
  subst; eauto; match goal with
  | [ H : typing nil _ (exp_var _) _ |- _ ] => 
    apply closed_no_var in H; inversion H
  end.
Qed.

Theorem canonical_forms_mu' : forall G t a A,
     typing G t a A 
  -> val a -> forall A', 
     (A = typ_mu A' \/ exists t', Boxed (typ_mu A') t' A)
  ->    (exists b, a = exp_roll b)
     \/ (exists n, a = exp_var n).
Proof.
  induction 1; intros; subst; eauto;
  val_inversion; destruct_things; subst; eauto;
  try solve [match goal with
  | [ H : _ = _ |- _ ] => inversion H
  end]; try match goal with
  | [ IH : _ -> forall _, _ -> ?P,
      H : Boxed _ _ _ |- ?P ] => 
    eapply IH; eauto; inversion H; subst; eauto
  end.
Qed.

Theorem canonical_forms_mu_closed : forall t a A,
     typing nil t a (typ_mu A) -> val a
  -> (exists b, a = exp_roll b).
Proof.
  intros t a A HTyp HVal;
  edestruct (canonical_forms_mu' _ _ _ _ HTyp); eauto;
  destruct_things;
  subst; eauto; match goal with
  | [ H : typing nil _ (exp_var _) _ |- _ ] => 
    apply closed_no_var in H; inversion H
  end.
Qed.

Theorem canonical_forms_sum' : forall G t a A,
     typing G t a A -> val a 
  -> forall B1 B2, 
     (A = typ_sum B1 B2 \/ exists t', Boxed (typ_sum B1 B2) t' A)
       ->    (exists a', a = exp_inl a')
          \/ (exists a', a = exp_inr a') 
          \/ (exists x, a = exp_var x).
Proof.
  induction 1; intros; subst; eauto;
  val_inversion; destruct_things; subst; eauto;
  try solve [match goal with
  | [ H : _ = _ |- _ ] => inversion H
  end]; try match goal with
  | [ IH : _ -> forall _ _, _ -> ?P,
      H : Boxed _ _ _ |- ?P ] => 
    eapply IH; eauto; inversion H; subst; eauto
  end.
Qed.

Theorem canonical_forms_sum_closed : forall t a B1 B2,
     typing nil t a (typ_sum B1 B2) -> val a
  -> (exists b, a = exp_inl b) \/ (exists b, a = exp_inr b).
Proof.
  intros t a B1 B2 HTyp HVal;
  edestruct (canonical_forms_sum' _ _ _ _ HTyp); eauto;
  destruct_things;
  subst; eauto; match goal with
  | [ H : typing nil _ (exp_var _) _ |- _ ] => 
    apply closed_no_var in H; inversion H
  end.
Qed.

Ltac go_go_canonical_forms :=
  match goal with
  | [ HT : typing nil _ ?a typ_nat, HVal : val ?a |- _ ] =>
    apply canonical_forms_nat_closed in HT; auto;
    destruct_things; subst
  | [ HT : typing nil _ ?a (typ_arr _ _), HVal : val ?a |- _ ] =>
    apply canonical_forms_arrow_closed in HT; auto;
    destruct_things; subst
  | [ HT : typing nil _ ?a (typ_mu _), HVal : val ?a |- _ ] =>
    apply canonical_forms_mu_closed in HT; auto;
    destruct_things; subst
  | [ HT : typing nil _ ?a (typ_sum _ _), HVal : val ?a |- _ ] =>
    apply canonical_forms_sum_closed in HT; auto;
    destruct_things; subst
  end.


Theorem progress' : forall G t a A,
     typing G t a A 
  -> G = nil -> val a \/ exists a', step a a'.
Proof.
  induction 1; intros;
  try match goal with
  | [ H : ?G1 ++ _ :: _ = nil |- _ ] => destruct G1; simpl in H; inversion H
  end; subst;
  repeat (match goal with
  | [ H : nil = nil -> _ \/ _ |- _ ] => destruct (H eq_refl); clear H
  | [ H : exists _, step ?a _ |- _ ] => 
    destruct H as [? ?]
  end);
  try solve [val_inversion; eauto;
  go_go_canonical_forms; val_inversion; eauto];
  match goal with
  | [ H : typing _ _ _ (typ_at (typ_sum _ _) _) |- _ ] =>
    edestruct (canonical_forms_sum' _ _ _ _ H); destruct_things; subst;
    val_inversion; eauto;
    match goal with
    | [ H' : typing nil _ (exp_var _) _ |- _ ] =>
      apply closed_no_var in H; inversion H
    end
  end.
Qed.

Theorem progress : forall t a A,
     typing nil t a A 
  -> val a \/ exists a', step a a'.
Proof.
  intros; eauto using progress'.
Qed.
